from __future__ import annotations
import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, GLib
import boto3

import string
import random
import time
import json
from copy import deepcopy
import logging

from .aws_s3_bucket_engine_advanced_settings import (
    AWSS3BucketEngineAdvancedSettings,
)
from .base_s3_bucket_engine import (
    BaseS3BucketEngine,
    BaseS3BucketEngineThread,
    AVAILABLE_CONFIGURATIONS,
)
from ..file import S3Object
from ..utils.decorators import (
    exported_filetype,
    with_advanced_settings,
    with_pango_docs,
)

logger = logging.getLogger(__name__)

ABORT_MESSAGE = "Ensure you are using valid credentials and bucket name, and that an appropriate policy is attached to the user or role."


@exported_filetype(filetype=S3Object)
@with_advanced_settings(
    engine_advanced_settings=AWSS3BucketEngineAdvancedSettings
)
@with_pango_docs(filename="aws_s3_bucket_engine.pango")
class AWSS3BucketEngine(BaseS3BucketEngine):

    NAME = "AWS S3 Bucket Monitor"

    def __init__(self, appwindow):
        super().__init__(appwindow, AWSS3BucketEngineThread, ABORT_MESSAGE)

        # Needs:
        # 1. region
        # 2. bucket name
        # 3. access key
        # 4. secret key

        # add bucket name -> this bucket must already exist!!!
        self.attach(
            Gtk.Label(
                label="Bucket Name",
                halign=Gtk.Align.START,
                valign=Gtk.Align.CENTER,
                hexpand=False,
                vexpand=False,
            ),
            0,
            0,
            1,
            1,
        )
        self._bucket_name_entry = self.register_widget(
            Gtk.Entry(
                halign=Gtk.Align.FILL,
                valign=Gtk.Align.CENTER,
                hexpand=True,
                vexpand=False,
            ),
            "bucket_name",
        )
        self.attach(self._bucket_name_entry, 1, 0, 1, 1)

        # Process existing files in monitored directory
        process_existing_files_switch = self.register_widget(
            Gtk.Switch(
                halign=Gtk.Align.CENTER,
                valign=Gtk.Align.CENTER,
                hexpand=False,
                vexpand=False,
                active=False,
            ),
            "process_existing_files",
        )
        self.attach(process_existing_files_switch, 2, 0, 1, 1)
        self.attach(
            Gtk.Label(
                label="Process existing files in bucket",
                halign=Gtk.Align.START,
                valign=Gtk.Align.CENTER,
                hexpand=True,
                vexpand=False,
            ),
            3,
            0,
            1,
            1,
        )

        # Access key
        self.attach(
            Gtk.Label(
                label="Access Key",
                halign=Gtk.Align.START,
                valign=Gtk.Align.CENTER,
                hexpand=False,
                vexpand=False,
            ),
            0,
            1,
            1,
            1,
        )
        self._access_key_entry = self.register_widget(
            Gtk.Entry(
                halign=Gtk.Align.FILL,
                valign=Gtk.Align.CENTER,
                hexpand=True,
                vexpand=False,
            ),
            "access_key",
        )
        self.attach(self._access_key_entry, 1, 1, 1, 1)

        # Secret key
        self.attach(
            Gtk.Label(
                label="Secret Key",
                halign=Gtk.Align.START,
                valign=Gtk.Align.CENTER,
                hexpand=False,
                vexpand=False,
            ),
            2,
            1,
            1,
            1,
        )
        self._secret_key_entry = self.register_widget(
            Gtk.Entry(
                visibility=False,
                halign=Gtk.Align.FILL,
                valign=Gtk.Align.CENTER,
                hexpand=True,
                vexpand=False,
            ),
            "secret_key",
            exportable=False,
        )
        self.attach(self._secret_key_entry, 3, 1, 1, 1)

        # connect signal handlers to determine validity
        self._access_key_entry.connect("changed", self._s3_entry_changed_cb)
        self._secret_key_entry.connect("changed", self._s3_entry_changed_cb)
        self._bucket_name_entry.connect("changed", self._s3_entry_changed_cb)

    def _s3_entry_changed_cb(self, entry):
        # todo: implement better bucket name validation
        if (
            self.params.bucket_name
            and self.params.access_key
            and self.params.secret_key
        ):
            self._valid = True
        else:
            self._valid = False

        self.notify("valid")

    def cleanup(self):
        # pylint: disable=no-member
        # we are going to do this 'ask for forgiveness, not permission'-style

        # restore old bucket notifications
        if hasattr(self, "old_bucket_notification_config"):
            try:
                self.s3_client.put_bucket_notification_configuration(
                    Bucket=self.params.bucket_name,
                    NotificationConfiguration=self.old_bucket_notification_config,
                )
            except Exception as e:
                logger.exception(
                    f"Could not restore bucket notification config: {str(e)}"
                )

        # delete SQS queue
        if hasattr(self, "queue_url"):
            try:
                self.sqs_client.delete_queue(QueueUrl=self.queue_url)
            except Exception as e:
                logger.exception(
                    f"Could not delete SQS queue {self.queue_url}: {str(e)}"
                )

        # delete SQS DLQ
        if hasattr(self, "dlq_url"):
            try:
                self.sqs_client.delete_queue(QueueUrl=self.dlq_url)
            except Exception as e:
                logger.exception(
                    f"Could not delete SQS queue {self.dlq_url}: {str(e)}"
                )

        super().cleanup()


class AWSS3BucketEngineThread(BaseS3BucketEngineThread):
    def get_full_name(self, key):
        return f"https://{self.params.bucket_name}.s3.{self._client_options['region_name']}.amazonaws.com/{key}"

    def run(self):
        self._client_options = self._engine._get_client_options()

        if not self.get_region_name("us-east-1"):
            return

        # set up proper s3 client
        self._engine.s3_client = boto3.client("s3", **self._client_options)

        # set up sqs client
        self._engine.sqs_client = boto3.client("sqs", **self._client_options)

        # create queue
        self._engine.queue_name = (
            "rfi-file-monitor-s3-bucket-engine-"
            + "".join(random.choice(string.ascii_lowercase) for i in range(6))
        )
        GLib.idle_add(
            self._task_window.set_text, "<b>Creating SQS queues...</b>"
        )
        try:
            self._engine.queue_url = self._engine.sqs_client.create_queue(
                QueueName=self._engine.queue_name
            )["QueueUrl"]
        except Exception as e:
            self._engine.cleanup()
            GLib.idle_add(
                self._engine.abort,
                self._task_window,
                e,
                priority=GLib.PRIORITY_HIGH,
            )
            return

        # create dead-letter-queue
        self._engine.dlq_name = self._engine.queue_name + "-dlq"
        try:
            self._engine.dlq_url = self._engine.sqs_client.create_queue(
                QueueName=self._engine.dlq_name
            )["QueueUrl"]
        except Exception as e:
            self._engine.cleanup()
            GLib.idle_add(
                self._engine.abort,
                self._task_window,
                e,
                priority=GLib.PRIORITY_HIGH,
            )
            return

        # sleep 1 second to make sure the queue is available
        time.sleep(1)

        try:
            # get queue ARN
            self._engine.queue_arn = (
                self._engine.sqs_client.get_queue_attributes(
                    QueueUrl=self._engine.queue_url, AttributeNames=["QueueArn"]
                )["Attributes"]["QueueArn"]
            )

            # get dlq ARN
            self._engine.dlq_arn = self._engine.sqs_client.get_queue_attributes(
                QueueUrl=self._engine.dlq_url, AttributeNames=["QueueArn"]
            )["Attributes"]["QueueArn"]
        except Exception as e:
            self._engine.cleanup()
            GLib.idle_add(
                self._engine.abort,
                self._task_window,
                e,
                priority=GLib.PRIORITY_HIGH,
            )
            return

        # set queue policy
        sqs_policy = {
            "Version": "2012-10-17",
            "Id": "example-ID",
            "Statement": [
                {
                    "Sid": "RFI-File-Monitor-SQS-ID",
                    "Effect": "Allow",
                    "Principal": {"AWS": "*"},
                    "Action": ["SQS:SendMessage"],
                    "Resource": self._engine.queue_arn,
                    "Condition": {
                        "ArnLike": {
                            "aws:SourceArn": f"arn:aws:s3:*:*:{self.params.bucket_name}"
                        },
                    },
                }
            ],
        }

        try:
            self._engine.sqs_client.set_queue_attributes(
                QueueUrl=self._engine.queue_url,
                Attributes={"Policy": json.dumps(sqs_policy)},
            )
        except Exception as e:
            self._engine.cleanup()
            GLib.idle_add(
                self._engine.abort,
                self._task_window,
                e,
                priority=GLib.PRIORITY_HIGH,
            )
            return

        # set dlq policy
        redrive_policy = {
            "deadLetterTargetArn": self._engine.dlq_arn,
            "maxReceiveCount": "10",
        }

        try:
            self._engine.sqs_client.set_queue_attributes(
                QueueUrl=self._engine.queue_url,
                Attributes={"RedrivePolicy": json.dumps(redrive_policy)},
            )
        except Exception as e:
            self._engine.cleanup()
            GLib.idle_add(
                self._engine.abort,
                self._task_window,
                e,
                priority=GLib.PRIORITY_HIGH,
            )
            return

        GLib.idle_add(
            self._task_window.set_text,
            "<b>Configuring bucket notifications...</b>",
        )
        try:
            # get current bucket notifications
            response = (
                self._engine.s3_client.get_bucket_notification_configuration(
                    Bucket=self.params.bucket_name,
                )
            )

            self._engine.old_bucket_notification_config = {
                configs: response.get(configs, [])
                for configs in AVAILABLE_CONFIGURATIONS
            }

            new_bucket_notification_config = deepcopy(
                self._engine.old_bucket_notification_config
            )

            sqs_notification = {
                "QueueArn": self._engine.queue_arn,
                "Events": [
                    "s3:ObjectCreated:*",
                    "s3:ObjectRemoved:*",
                    "s3:ObjectRestore:*",
                ],
            }

            new_bucket_notification_config["QueueConfigurations"].append(
                sqs_notification
            )

            # set bucket notifications
            self._engine.s3_client.put_bucket_notification_configuration(
                Bucket=self.params.bucket_name,
                NotificationConfiguration=new_bucket_notification_config,
            )
        except Exception as e:
            self._engine.cleanup()
            GLib.idle_add(
                self._engine.abort,
                self._task_window,
                e,
                priority=GLib.PRIORITY_HIGH,
            )
            return

        # if required, add existing files to queue
        if (
            self.params.process_existing_files
            and not self.process_existing_files()
        ):
            return

        # if we get here, things should be working.
        # close task_window
        GLib.idle_add(
            self._engine.kill_task_window,
            self._task_window,
            priority=GLib.PRIORITY_HIGH,
        )

        # start the big while loop and start consuming incoming messages
        while True:

            if self._should_exit:
                self._engine.cleanup()
                return

            try:
                resp = self._engine.sqs_client.receive_message(
                    QueueUrl=self._engine.queue_url,
                    AttributeNames=["All"],
                    MaxNumberOfMessages=10,
                    WaitTimeSeconds=10,
                )
            except Exception as e:
                self._engine.cleanup()
                GLib.idle_add(
                    self._engine.abort, None, e, priority=GLib.PRIORITY_HIGH
                )
                return

            if "Messages" not in resp:
                continue

            for message in resp["Messages"]:
                body = json.loads(message["Body"])
                # we are going to assume 1 record per message
                try:
                    record = body["Records"][0]
                    event_name = record["eventName"]
                except Exception as e:
                    logger.info(f"Ignoring {message=} because of {str(e)}")
                    continue

                if event_name.startswith("ObjectCreated"):
                    # new file created!
                    s3_info = record["s3"]
                    if not self.process_new_file(s3_info):
                        continue

            # delete messages from the queue
            entries = [
                {"Id": msg["MessageId"], "ReceiptHandle": msg["ReceiptHandle"]}
                for msg in resp["Messages"]
            ]

            try:
                resp = self._engine.sqs_client.delete_message_batch(
                    QueueUrl=self._engine.queue_url, Entries=entries
                )
            except Exception as e:
                self._engine.cleanup()
                GLib.idle_add(
                    self._engine.abort, None, e, priority=GLib.PRIORITY_HIGH
                )
                return

            if len(resp["Successful"]) != len(entries):
                logger.warn(
                    f"Failed to delete messages: entries={entries!r} resp={resp!r}"
                )



from abc import ABC, abstractmethod, ABCMeta
import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk

from .file import File
from .utils.widgetparams import WidgetParams


#
# This is my attempt at extending Gtk.Frame with
# abstract methods to turn it into an abstract class.
# Though this seems to be done properly, classes derived from
# the Operation abstract class do not appear to cause errors
# at runtime if abstract methods were not implemented.
# So, the errors that are produced and bring the app down are
# NotImplementedError exceptions...
#
class OperationMeta(ABCMeta, Gtk.Frame.__class__):
    pass


class Operation(ABC, Gtk.Frame, WidgetParams, metaclass=OperationMeta):
    @abstractmethod
    def __init__(self, *args, **kwargs):
        self._appwindow = kwargs.pop("appwindow")
        self._index: int = 0

        self._label = Gtk.Label(
            halign=Gtk.Align.START,
            valign=Gtk.Align.CENTER,
            hexpand=False,
            vexpand=False,
            label=f"Operation without index: {self.NAME}",
        )

        label_grid = Gtk.Grid(
            halign=Gtk.Align.START,
            valign=Gtk.Align.CENTER,
            hexpand=False,
            vexpand=False,
            column_spacing=5,
            border_width=5,
        )
        label_grid.attach(self._label, 0, 0, 1, 1)

        # add delete button
        delete_button = Gtk.Button(
            image=Gtk.Image(
                icon_name="edit-delete-symbolic",
                icon_size=Gtk.IconSize.SMALL_TOOLBAR,
            ),
            halign=Gtk.Align.START,
            valign=Gtk.Align.CENTER,
            hexpand=False,
            vexpand=False,
        )
        label_grid.attach(delete_button, 1, 0, 1, 1)

        delete_button.connect("clicked", self._delete_clicked_cb)

        if (
            type(self)
            in self._appwindow.get_property("application").pango_docs_map
        ):
            help_button = Gtk.Button(
                image=Gtk.Image(
                    icon_name="dialog-question-symbolic",
                    icon_size=Gtk.IconSize.SMALL_TOOLBAR,
                ),
                halign=Gtk.Align.START,
                valign=Gtk.Align.CENTER,
                hexpand=False,
                vexpand=False,
            )
            label_grid.attach(help_button, 2, 0, 1, 1)

            help_button.connect("clicked", self._help_clicked_cb)

        kwargs.update(
            dict(
                label_widget=label_grid,
                halign=Gtk.Align.FILL,
                valign=Gtk.Align.FILL,
                hexpand=True,
                vexpand=True,
                border_width=5,
                label_xalign=0.5,
            )
        )
        Gtk.Frame.__init__(self, *args, **kwargs)
        WidgetParams.__init__(self)

    def _delete_clicked_cb(self, button):
        self._appwindow._remove_operation(self)

    def _help_clicked_cb(self, button):
        dialog = self._appwindow.get_property("application").help_window
        dialog.props.transient_for = self._appwindow
        dialog.select_item(type(self))
        dialog.present()

    @property
    def appwindow(self):
        """
        Returns the ApplicationWindow instance this operation is a part of
        """
        return self._appwindow

    @property
    def index(self) -> int:
        """
        The position of the operation in the list of operations in the applicationwindow
        """
        return self._index

    @index.setter
    def index(self, value: int):
        self._index = value
        self._label.props.label = f"Operation {self._index + 1}: {self.NAME}"

    @property
    @classmethod
    @abstractmethod
    def NAME(cls) -> str:
        """
        A descriptive short name for the operation,
        that will show up in the list of available operations,
        and will be used in the title of this frame.
        """
        raise NotImplementedError

    @abstractmethod
    def run(self, file: File):
        """
        This method will process File and execute the operation.
        Since this will be done in a worker-thread, you cannot directly
        call any methods that update the GUI.

        """
        raise NotImplementedError

    def preflight_check(self):
        """
        This method will be used to check that all widgets have valid information,
        and should also be used to extract that information into variables that
        can be used in the run() method, meaning that no Gtk calls will need to be made
        when calling run() afterwards.
        Feel free to use it also for initializing other stuff, such as opening connections.

        If something goes wrong during this method, an exception will be raised.
        """
        return

    def preflight_cleanup(self, success: bool):
        """
        Use this method to do some cleanup after the preflight_check.
        The value of success is determined by the global outcome of the
        preflight_check, not just the current operation.
        """
        return

    def destroy_operation(self):
        """
        use this method to remove widgets from the GUI and widget registry
        """
        return



import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk
import paramiko
from munch import Munch
from tenacity import (
    retry,
    stop_after_attempt,
    wait_exponential,
    before_log,
    after_log,
    before_sleep_log,
)

from ..operation import Operation
from ..utils.exceptions import SkippedOperation
from ..file import File, RegularFile, Directory
from ..utils import monitor_retry_condition
from ..utils.decorators import (
    with_pango_docs,
    supported_filetypes,
    add_directory_support,
)

import logging
import os
import tempfile
from pathlib import PurePosixPath, Path
import stat
import posixpath
from typing import List
from threading import RLock

logger = logging.getLogger(__name__)


@with_pango_docs(filename="sftp_uploader.pango")
@supported_filetypes(filetypes=(RegularFile, Directory))
class SftpUploaderOperation(Operation):

    NAME = "SFTP Uploader"

    def __init__(self, *args, **kwargs):
        Operation.__init__(self, *args, **kwargs)
        self._grid = Gtk.Grid(
            border_width=5,
            row_spacing=5,
            column_spacing=5,
            halign=Gtk.Align.FILL,
            valign=Gtk.Align.CENTER,
            hexpand=True,
            vexpand=False,
        )
        self.add(self._grid)

        # we need boxes for at least
        # hostname
        # port
        # username
        # password
        # auto add new host keys
        # remote directory
        # create remote directory if necessary

        # Hostname
        tempgrid = Gtk.Grid(
            row_spacing=5,
            column_spacing=5,
            halign=Gtk.Align.FILL,
            valign=Gtk.Align.CENTER,
            hexpand=True,
            vexpand=False,
        )
        self._grid.attach(tempgrid, 0, 0, 1, 1)
        tempgrid.attach(
            Gtk.Label(
                label="Hostname",
                halign=Gtk.Align.START,
                valign=Gtk.Align.CENTER,
                hexpand=False,
                vexpand=False,
            ),
            0,
            0,
            1,
            1,
        )
        widget = self.register_widget(
            Gtk.Entry(
                halign=Gtk.Align.FILL,
                valign=Gtk.Align.CENTER,
                hexpand=True,
                vexpand=False,
            ),
            "hostname",
        )
        tempgrid.attach(widget, 1, 0, 1, 1)

        # Port
        tempgrid.attach(
            Gtk.Label(
                label="Port",
                halign=Gtk.Align.CENTER,
                valign=Gtk.Align.CENTER,
                hexpand=False,
                vexpand=False,
            ),
            2,
            0,
            1,
            1,
        )
        widget = self.register_widget(
            Gtk.SpinButton(
                adjustment=Gtk.Adjustment(
                    lower=1, upper=10000, value=5, page_size=0, step_increment=1
                ),
                value=22,
                update_policy=Gtk.SpinButtonUpdatePolicy.IF_VALID,
                numeric=True,
                climb_rate=5,
                halign=Gtk.Align.CENTER,
                valign=Gtk.Align.CENTER,
                hexpand=False,
                vexpand=False,
            ),
            "port",
        )
        tempgrid.attach(widget, 3, 0, 1, 1)

        # Username
        tempgrid = Gtk.Grid(
            row_spacing=5,
            column_spacing=5,
            halign=Gtk.Align.FILL,
            valign=Gtk.Align.CENTER,
            hexpand=True,
            vexpand=False,
        )
        self._grid.attach(tempgrid, 0, 1, 1, 1)
        tempgrid.attach(
            Gtk.Label(
                label="Username",
                halign=Gtk.Align.CENTER,
                valign=Gtk.Align.CENTER,
                hexpand=False,
                vexpand=False,
            ),
            0,
            0,
            1,
            1,
        )
        widget = self.register_widget(
            Gtk.Entry(
                halign=Gtk.Align.FILL,
                valign=Gtk.Align.CENTER,
                hexpand=True,
                vexpand=False,
            ),
            "username",
        )
        tempgrid.attach(widget, 1, 0, 1, 1)

        # Password
        tempgrid.attach(
            Gtk.Label(
                label="Password",
                halign=Gtk.Align.CENTER,
                valign=Gtk.Align.CENTER,
                hexpand=False,
                vexpand=False,
            ),
            2,
            0,
            1,
            1,
        )
        widget = self.register_widget(
            Gtk.Entry(
                visibility=False,
                halign=Gtk.Align.FILL,
                valign=Gtk.Align.CENTER,
                hexpand=True,
                vexpand=False,
            ),
            "password",
            exportable=False,
        )
        tempgrid.attach(widget, 3, 0, 1, 1)

        # Remote directory
        tempgrid = Gtk.Grid(
            row_spacing=5,
            column_spacing=5,
            halign=Gtk.Align.FILL,
            valign=Gtk.Align.CENTER,
            hexpand=True,
            vexpand=False,
        )
        self._grid.attach(tempgrid, 0, 2, 1, 1)
        tempgrid.attach(
            Gtk.Label(
                label="Destination folder",
                halign=Gtk.Align.CENTER,
                valign=Gtk.Align.CENTER,
                hexpand=False,
                vexpand=False,
            ),
            0,
            0,
            1,
            1,
        )
        widget = self.register_widget(
            Gtk.Entry(
                halign=Gtk.Align.FILL,
                valign=Gtk.Align.CENTER,
                hexpand=True,
                vexpand=False,
            ),
            "destination",
        )
        tempgrid.attach(widget, 1, 0, 1, 1)

        # Create directory
        widget = self.register_widget(
            Gtk.CheckButton(
                active=True,
                label="Create destination folder if necessary",
                halign=Gtk.Align.END,
                valign=Gtk.Align.CENTER,
                hexpand=False,
                vexpand=False,
            ),
            "force_folder_creation",
        )
        tempgrid.attach(widget, 2, 0, 1, 1)

        # Automatically accept new host keys
        tempgrid = Gtk.Grid(
            row_spacing=5,
            column_spacing=5,
            halign=Gtk.Align.FILL,
            valign=Gtk.Align.CENTER,
            hexpand=True,
            vexpand=False,
        )
        self._grid.attach(tempgrid, 0, 3, 1, 1)
        widget = self.register_widget(
            Gtk.CheckButton(
                active=False,
                label="Automatically accept new host keys (dangerous!!)",
                halign=Gtk.Align.END,
                valign=Gtk.Align.CENTER,
                hexpand=False,
                vexpand=False,
            ),
            "auto_add_keys",
        )
        tempgrid.attach(widget, 0, 0, 1, 1)

        # Advanced options expander
        advanced_options = Gtk.Expander(
            label="Advanced Options",
            halign=Gtk.Align.FILL,
            valign=Gtk.Align.CENTER,
            hexpand=True,
            vexpand=False,
        )
        self._grid.attach(advanced_options, 0, 4, 1, 1)

        advanced_options_grid = Gtk.Grid(
            halign=Gtk.Align.FILL,
            valign=Gtk.Align.CENTER,
            hexpand=True,
            vexpand=False,
            row_spacing=5,
            column_spacing=5,
        )
        advanced_options.add(advanced_options_grid)

        advanced_options_grid.attach(
            self._get_chmod_grid("file", "644"), 0, 0, 1, 1
        )
        advanced_options_grid.attach(
            self._get_chmod_grid("directory", "755"), 1, 0, 1, 1
        )

    def _get_chmod_grid(self, kind: str, default_octal: str):
        permissions = (
            (stat.S_IRUSR, stat.S_IWUSR, stat.S_IXUSR),
            (stat.S_IRGRP, stat.S_IWGRP, stat.S_IXGRP),
            (stat.S_IROTH, stat.S_IWOTH, stat.S_IXOTH),
        )

        default_octal_as_int = int(default_octal, base=8)

        chmod_grid = Gtk.Grid(
            halign=Gtk.Align.FILL,
            valign=Gtk.Align.CENTER,
            hexpand=True,
            vexpand=False,
            row_spacing=5,
            column_spacing=5,
        )

        title_grid = Gtk.Grid(
            halign=Gtk.Align.FILL,
            valign=Gtk.Align.CENTER,
            hexpand=True,
            vexpand=False,
            row_spacing=5,
            column_spacing=5,
        )
        chmod_grid.attach(title_grid, 0, 0, 4, 1)
        chmod_grid._checkbutton = self.register_widget(
            Gtk.CheckButton(
                label=f"<b>Override {kind} UNIX permissions</b>",
                halign=Gtk.Align.START,
                valign=Gtk.Align.CENTER,
                hexpand=False,
                vexpand=False,
            ),
            f"{kind}_chmod_enabled",
        )
        chmod_grid._checkbutton.get_child().props.use_markup = True
        title_grid.attach(chmod_grid._checkbutton, 0, 0, 1, 1)

        chmod_grid._entry = self.register_widget(
            Gtk.Entry(
                editable=False,
                text=default_octal,
                halign=Gtk.Align.START,
                valign=Gtk.Align.CENTER,
                hexpand=False,
                vexpand=False,
            ),
            f"{kind}_chmod_octal",
        )
        chmod_grid._entry_handler_id = chmod_grid._entry.connect(
            "changed", self._chmod_grid_entry_changed_cb, chmod_grid
        )
        title_grid.attach(chmod_grid._entry, 1, 0, 1, 1)

        label = Gtk.Label(
            label="<b>Read</b>",
            use_markup=True,
            halign=Gtk.Align.CENTER,
            valign=Gtk.Align.CENTER,
            hexpand=True,
            vexpand=False,
        )
        chmod_grid.attach(label, 1, 1, 1, 1)
        label = Gtk.Label(
            label="<b>Write</b>",
            use_markup=True,
            halign=Gtk.Align.CENTER,
            valign=Gtk.Align.CENTER,
            hexpand=True,
            vexpand=False,
        )
        chmod_grid.attach(label, 2, 1, 1, 1)
        label = Gtk.Label(
            label="<b>Execute</b>",
            use_markup=True,
            halign=Gtk.Align.CENTER,
            valign=Gtk.Align.CENTER,
            hexpand=True,
            vexpand=False,
        )
        chmod_grid.attach(label, 3, 1, 1, 1)

        label = Gtk.Label(
            label="<b>User</b>",
            use_markup=True,
            halign=Gtk.Align.CENTER,
            valign=Gtk.Align.CENTER,
            hexpand=True,
            vexpand=False,
        )
        chmod_grid.attach(label, 0, 2, 1, 1)
        label = Gtk.Label(
            label="<b>Group</b>",
            use_markup=True,
            halign=Gtk.Align.CENTER,
            valign=Gtk.Align.CENTER,
            hexpand=True,
            vexpand=False,
        )
        chmod_grid.attach(label, 0, 3, 1, 1)
        label = Gtk.Label(
            label="<b>Others</b>",
            use_markup=True,
            halign=Gtk.Align.CENTER,
            valign=Gtk.Align.CENTER,
            hexpand=True,
            vexpand=False,
        )
        chmod_grid.attach(label, 0, 4, 1, 1)

        for i in range(3):
            for j in range(3):
                check_button = Gtk.CheckButton(
                    halign=Gtk.Align.CENTER,
                    valign=Gtk.Align.CENTER,
                    hexpand=True,
                    vexpand=False,
                )
                chmod_grid.attach(check_button, j + 1, i + 2, 1, 1)
                check_button._permission = permissions[i][j]
                check_button._handler_id = check_button.connect(
                    "clicked",
                    self._chmod_grid_check_button_clicked_cb,
                    chmod_grid,
                )
                if check_button._permission & default_octal_as_int:
                    with check_button.handler_block(check_button._handler_id):
                        check_button.props.active = True

        return chmod_grid

    def _chmod_grid_check_button_clicked_cb(self, button, chmod_grid):
        # get entry value as int
        octal = int(chmod_grid._entry.props.text.strip(), base=8)

        if button.props.active:
            octal |= button._permission
        else:
            octal &= ~button._permission

        # update entry
        with chmod_grid._entry.handler_block(chmod_grid._entry_handler_id):
            chmod_grid._entry.props.text = f"{octal:03o}"

    def _chmod_grid_entry_changed_cb(self, entry, chmod_grid):
        # this method will only be called when loading from yaml
        # get entry value as int
        octal = int(chmod_grid._entry.props.text.strip(), base=8)

        for child in chmod_grid:
            if hasattr(child, "_permission"):
                # block the clicked handler to avoid signal loop
                with child.handler_block(child._handler_id):
                    child.props.active = octal & child._permission

    @classmethod
    def _preflight_check(cls, params: Munch):
        # try connecting to server and copy a simple file
        with paramiko.Transport(
            (params.hostname, int(params.port))
        ) as transport:

            transport.connect(
                username=params.username,
                password=params.password,
            )
            folder_created = False
            with paramiko.SFTPClient.from_transport(transport) as sftp_client:
                try:
                    sftp_client.chdir(params.destination)
                except IOError:
                    if params.force_folder_creation:
                        makedirs(sftp_client, params.destination)
                        if params.directory_chmod_enabled:
                            sftp_client.chmod(
                                params.destination,
                                int(params.directory_chmod_octal, base=8),
                            )
                        folder_created = True
                    else:
                        raise
                # cd back to home folder
                sftp_client.chdir()

                # try copying a file
                with tempfile.NamedTemporaryFile(delete=False) as f:
                    f.write(os.urandom(1024))  # 1 kB
                    tmpfile = f.name
                file_destination = (
                    params.destination + "/" + os.path.basename(tmpfile)
                )
                try:
                    sftp_client.put(tmpfile, file_destination)
                except:
                    if folder_created:
                        sftp_client.rmdir(params.destination)
                    raise
                else:
                    try:
                        if params.file_chmod_enabled:
                            sftp_client.chmod(
                                file_destination,
                                int(params.file_chmod_octal, base=8),
                            )
                    except:
                        # cleanup
                        sftp_client.remove(file_destination)
                        if folder_created:
                            sftp_client.rmdir(params.destination)
                        raise
                    else:
                        # if successful, remove it
                        sftp_client.remove(file_destination)
                finally:
                    os.unlink(tmpfile)

    def preflight_check(self):
        self._processed_dirs: List[str] = []
        self._processed_dirs_lock: RLock = RLock()
        self._preflight_check(self.params)

    @classmethod
    def _attach_metadata(
        cls,
        file: File,
        remote_filename_full: str,
        params: Munch,
        operation_index: int,
    ):
        file.operation_metadata[operation_index] = {
            "sftp url": f"sftp://{params.username}@{params.hostname}:{int(params.port)}{remote_filename_full}"
        }

    @classmethod
    @retry(
        retry=monitor_retry_condition(),
        reraise=True,
        stop=stop_after_attempt(5),
        wait=wait_exponential(),
        before=before_log(logger, logging.DEBUG),
        after=after_log(logger, logging.DEBUG),
        before_sleep=before_sleep_log(logger, logging.DEBUG),
    )
    def _run(
        cls,
        file: File,
        params: Munch,
        operation_index: int,
        processed_dirs: List[str],
        processed_dirs_lock: RLock,
    ):
        try:
            with paramiko.Transport(
                (params.hostname, int(params.port))
            ) as transport:

                transport.connect(
                    username=params.username,
                    password=params.password,
                )

                with paramiko.SFTPClient.from_transport(
                    transport
                ) as sftp_client:
                    sftp_client.chdir(params.destination)
                    rel_filename = str(
                        PurePosixPath(*file.relative_filename.parts)
                    )
                    dirname = posixpath.dirname(rel_filename)
                    with processed_dirs_lock:
                        if dirname not in processed_dirs:
                            if not isdir(sftp_client, dirname):
                                makedirs(sftp_client, dirname)

                                if params.directory_chmod_enabled:
                                    sftp_client.chmod(
                                        dirname,
                                        int(
                                            params.directory_chmod_octal, base=8
                                        ),
                                    )

                            processed_dirs.append(dirname)

                    # check if file already exists
                    # Note: ideally this would be done by calculating the checksum
                    # on the server, but very few SSH server implementations support
                    # this protocol extension, even though Paramiko does:
                    # http://docs.paramiko.org/en/stable/api/sftp.html#paramiko.sftp_file.SFTPFile.check
                    try:
                        remote_stat = sftp_client.stat(rel_filename)
                    except IOError:
                        pass
                    else:
                        # file exists -> compare with local stat
                        local_stat = Path(file.filename).stat()
                        # if local file is more recent or size differs -> upload again!
                        if (
                            local_stat.st_size == remote_stat.st_size
                            and local_stat.st_mtime <= remote_stat.st_mtime
                        ):
                            # add object URL to metadata
                            remote_filename_full = sftp_client.normalize(
                                rel_filename
                            )
                            cls._attach_metadata(
                                file,
                                remote_filename_full,
                                params,
                                operation_index,
                            )
                            raise SkippedOperation(
                                "File has been uploaded already"
                            )

                    # upload the file to the remote server
                    sftp_client.put(
                        file.filename,
                        rel_filename,
                        callback=SftpProgressPercentage(file, operation_index),
                    )
                    if params.file_chmod_enabled:
                        sftp_client.chmod(
                            rel_filename, int(params.file_chmod_octal, base=8)
                        )
                    remote_filename_full = sftp_client.normalize(rel_filename)
        except SkippedOperation:
            raise
        except Exception as e:
            logger.exception(f"SftpUploaderOperation.run exception")
            return str(e)
        else:
            # add object URL to metadata
            cls._attach_metadata(
                file, remote_filename_full, params, operation_index
            )
        return None

    @add_directory_support
    def run(self, file: File):
        return self._run(
            file,
            self.params,
            self.index,
            self._processed_dirs,
            self._processed_dirs_lock,
        )


class SftpProgressPercentage(object):
    def __init__(self, file: File, operation_index: int):
        self._file = file
        self._last_percentage = 0
        self._operation_index = operation_index

    def __call__(self, bytes_so_far: int, bytes_total: int):
        percentage = (bytes_so_far / bytes_total) * 100
        if int(percentage) > self._last_percentage:
            self._last_percentage = int(percentage)
            self._file.update_progressbar(
                self._operation_index, self._last_percentage
            )


# the following methods have been inspired by pysftp
def isdir(sftp_client: paramiko.SFTPClient, remotepath: str):
    try:
        return stat.S_ISDIR(sftp_client.stat(remotepath).st_mode)
    except IOError:  # no such file
        return False


def isfile(sftp_client: paramiko.SFTPClient, remotepath: str):
    try:
        return stat.S_ISREG(sftp_client.stat(remotepath).st_mode)
    except IOError:  # no such file
        return False


def makedirs(sftp_client: paramiko.SFTPClient, remotedir: str, mode="777"):
    if isdir(sftp_client, remotedir):
        pass
    elif isfile(sftp_client, remotedir):
        raise OSError(
            f"a file with the same name as the remotedir {remotedir} already exists"
        )
    else:
        head, tail = posixpath.split(remotedir)
        if head and not isdir(sftp_client, head):
            makedirs(sftp_client, head, mode)
        if tail:
            sftp_client.mkdir(remotedir, mode=int(mode, 8))



import logging
import sys
import os
from pathlib import PurePath

import bugsnag
from bugsnag.handlers import BugsnagHandler
from .version import __version__

# set AWS_DATA_PATH to allow using Ceph specific boto3 API
os.environ["AWS_DATA_PATH"] = str(
    PurePath(__file__).parent.joinpath("data", "models")
)

BUGSNAG_API_KEY = "b19e59eb84b9eb30d31d57a97e03406a"

logger = logging.getLogger(__name__)

bugsnag.configure(
    api_key=BUGSNAG_API_KEY,
    app_version=__version__,
    auto_notify=True,
    auto_capture_sessions=True,
    notify_release_stages=["production"],
)

# main entrypoint
def main():
    from .application import Application

    # set up logging
    monitor_logger = logging.getLogger("rfi_file_monitor")
    monitor_logger.setLevel(logging.DEBUG)

    log_fmt_long = logging.Formatter(
        fmt="%(asctime)s %(name)s %(levelname)s: %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )

    # log to stdout
    log_handler_stream = logging.StreamHandler(sys.stdout)
    log_handler_stream.setFormatter(log_fmt_long)
    log_handler_stream.setLevel(logging.DEBUG)
    monitor_logger.addHandler(log_handler_stream)

    # log to bugsnag
    log_handler_bugsnag = BugsnagHandler()
    log_handler_bugsnag.setLevel(logging.WARNING)
    monitor_logger.addHandler(log_handler_bugsnag)

    app = Application()
    rv = app.run(sys.argv)
    sys.exit(rv)



import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk
from pathlib import Path
from typing import Final, Any, List, final, Optional

from munch import Munch, munchify
import logging

logger = logging.getLogger(__name__)


class WidgetParams:
    """
    Inheriting from this class
    """

    # pylint: disable=unsubscriptable-object
    def __init__(self, params: Optional[Munch] = None):
        if params:
            self._params = params
        else:
            self._params: Final[Munch[str, Any]] = Munch()

        self._signal_ids: Final[Munch[str, int]] = Munch()
        self._widgets: Final[Munch[str, Gtk.Widget]] = Munch()
        self._exportable_params: Final[List[str]] = list()
        self._desensitized_widgets: Final[List[Gtk.Widget]] = list()

    @final
    def set_sensitive(self, sensitive: bool):
        for widget in self.widgets.values():
            if widget not in self._desensitized_widgets:
                widget.set_sensitive(sensitive)

    @final
    def _entry_changed_cb(self, entry: Gtk.Entry, param_name: str):
        self._params[param_name] = entry.get_text().strip()

    @final
    def _checkbutton_toggled_cb(
        self, checkbutton: Gtk.CheckButton, param_name: str
    ):
        self._params[param_name] = checkbutton.get_active()

    @final
    def _filechooserbutton_selection_changed_cb(
        self, filechooserbutton: Gtk.FileChooserButton, param_name: str
    ):
        self._params[param_name] = filechooserbutton.get_filename()

    @final
    def _spinbutton_value_changed_cb(
        self, spinbutton: Gtk.SpinButton, param_name: str
    ):
        self._params[param_name] = spinbutton.get_value()

    @final
    def _combobox_changed_cb(self, combobox: Gtk.ComboBoxText, param_name: str):
        self._params[param_name] = combobox.get_active_text()

    @final
    def _switch_value_changed_cb(self, switch: Gtk.Switch, _, param_name: str):
        self._params[param_name] = switch.get_active()

    @final
    def register_widget(
        self,
        widget: Gtk.Widget,
        param_name: str,
        exportable: bool = True,
        desensitized: bool = False,
    ):

        if param_name in self._params:
            raise ValueError(
                "register_widget cannot overwrite existing parameters!"
            )

        if isinstance(widget, Gtk.Switch):
            self._params[param_name] = widget.get_active()
            self._signal_ids[param_name] = widget.connect(
                "notify::active", self._switch_value_changed_cb, param_name
            )
        elif isinstance(widget, Gtk.SpinButton):
            self._params[param_name] = widget.get_value()
            self._signal_ids[param_name] = widget.connect(
                "value-changed", self._spinbutton_value_changed_cb, param_name
            )
        elif isinstance(widget, Gtk.CheckButton):
            self._params[param_name] = widget.get_active()
            self._signal_ids[param_name] = widget.connect(
                "toggled", self._checkbutton_toggled_cb, param_name
            )
        elif isinstance(widget, Gtk.FileChooserButton):
            self._params[param_name] = widget.get_filename()
            self._signal_ids[param_name] = widget.connect(
                "selection-changed",
                self._filechooserbutton_selection_changed_cb,
                param_name,
            )
        elif isinstance(widget, Gtk.Entry):
            self._params[param_name] = widget.get_text().strip()
            self._signal_ids[param_name] = widget.connect(
                "changed", self._entry_changed_cb, param_name
            )
        elif isinstance(widget, Gtk.ComboBoxText):
            self._params[param_name] = widget.get_active_text()
            self._signal_ids[param_name] = widget.connect(
                "changed", self._combobox_changed_cb, param_name
            )
        else:
            raise NotImplementedError(
                f"register_widget: no support for {type(widget).__name__}"
            )

        self._widgets[param_name] = widget

        if exportable:
            self._exportable_params.append(param_name)

        if desensitized:
            self._desensitized_widgets.append(widget)

        return widget

    def update_from_dict(self, yaml_dict: dict):
        for param_name, value in yaml_dict.items():
            if param_name not in self._params:
                logger.warning(
                    f"update_from_dict: {param_name} not found in widget params!"
                )
                continue

            widget = self._widgets[param_name]

            with widget.handler_block(self._signal_ids[param_name]):
                self._params[param_name] = value

                if isinstance(widget, Gtk.SpinButton):
                    widget.set_value(value)
                elif isinstance(widget, Gtk.CheckButton) or isinstance(
                    widget, Gtk.Switch
                ):
                    widget.set_active(value)
                elif isinstance(widget, Gtk.FileChooserButton):
                    if value is None or not Path(value).exists:
                        continue
                    widget.set_filename(value)
                elif isinstance(widget, Gtk.Entry):
                    widget.set_text(value)
                elif isinstance(widget, Gtk.ComboBoxText):
                    for row in widget.get_model():
                        if row[0] == value:
                            widget.set_active_iter(row.iter)
                            break
                else:
                    raise NotImplementedError(
                        f"update_from_dict: no support for {type(widget).__name__}"
                    )

    @property
    def params(self) -> Munch:
        """
        A Munch dict containing the parameters that will be used by run
        """
        return self._params

    @property
    def exportable_params(self) -> Munch:
        """
        A Munch dict containing those parameters that have been considered safe for exporting.
        This will typically exclude widgets meant to hold passwords and other secrets.
        """
        return munchify(
            {param: self._params[param] for param in self._exportable_params}
        )

    @property
    def widgets(self) -> Munch:
        """
        A Munch dict containing the registered widgets
        """
        return self._widgets



from abc import ABC, abstractmethod, ABCMeta
import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk

from .utils import EXPAND_AND_FILL
from .engine import Engine


class EngineAdvancedSettingsMeta(ABCMeta, type(Gtk.Grid)):
    pass


class EngineAdvancedSettings(
    ABC, Gtk.Grid, metaclass=EngineAdvancedSettingsMeta
):
    @abstractmethod
    def __init__(self, engine: Engine):
        kwargs = dict(
            **EXPAND_AND_FILL,
            border_width=5,
            row_spacing=5,
            column_spacing=5,
        )
        Gtk.Grid.__init__(self, **kwargs)
        self._engine: Engine = engine



from __future__ import annotations

import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk
import boto3
import botocore
from munch import Munch
from tenacity import (
    retry,
    stop_after_attempt,
    wait_exponential,
    before_log,
    after_log,
    before_sleep_log,
)

from ..operation import Operation
from ..utils.exceptions import SkippedOperation
from ..file import File, RegularFile, Directory
from ..utils import query_metadata, monitor_retry_condition
from ..utils.decorators import (
    with_pango_docs,
    supported_filetypes,
    add_directory_support,
)
from ..utils.s3 import S3ProgressPercentage, TransferConfig, calculate_etag

import os
import logging
import tempfile
from pathlib import PurePosixPath, Path
import urllib
from typing import Sequence, Dict, Any, Optional

logger = logging.getLogger(__name__)

# useful info from help(boto3.session.Session.client)

ALLOWED_BUCKET_ACL_OPTIONS = (
    "ACL",
    "AccessControlPolicy",
    "GrantFullControl",
    "GrantRead",
    "GrantReadACP",
    "GrantWrite",
    "GrantWriteACP",
)

ALLOWED_OBJECT_ACL_OPTIONS = (
    "ACL",
    "AccessControlPolicy",
    "GrantFullControl",
    "GrantRead",
    "GrantReadACP",
    "GrantWrite",
    "GrantWriteACP",
    "RequestPayer",
    "VersionID",
)

AWS_S3_ENGINE_IGNORE_ME = "rfi-file-monitor-ignore-me"


@with_pango_docs(filename="s3_uploader.pango")
@supported_filetypes(filetypes=(RegularFile, Directory))
class S3UploaderOperation(Operation):
    NAME = "S3 Uploader"

    def __init__(self, *args, **kwargs):
        Operation.__init__(self, *args, **kwargs)
        self._grid = Gtk.Grid(
            border_width=5,
            row_spacing=5,
            column_spacing=5,
            halign=Gtk.Align.FILL,
            valign=Gtk.Align.CENTER,
            hexpand=True,
            vexpand=False,
        )
        self.add(self._grid)

        # we need boxes for at least:
        # * hostname (+ allow disabling security)
        # * bucket name (+ allow creating it)
        # * access
        # * secret

        # Endpoint
        self._grid.attach(
            Gtk.Label(
                label="Endpoint",
                halign=Gtk.Align.START,
                valign=Gtk.Align.CENTER,
                hexpand=False,
                vexpand=False,
            ),
            0,
            0,
            1,
            1,
        )
        self._endpoint_entry = self.register_widget(
            Gtk.Entry(
                text="https://s3.amazonaws.com",
                halign=Gtk.Align.FILL,
                valign=Gtk.Align.CENTER,
                hexpand=True,
                vexpand=False,
            ),
            "hostname",
        )
        self._grid.attach(self._endpoint_entry, 1, 0, 1, 1)
        widget = self.register_widget(
            Gtk.CheckButton(
                active=True,
                label="Verify SSL Certificates",
                halign=Gtk.Align.START,
                valign=Gtk.Align.CENTER,
                hexpand=False,
                vexpand=False,
            ),
            "hostname_ssl_verify",
        )
        self._grid.attach(widget, 2, 0, 1, 1)

        # Access key
        self._grid.attach(
            Gtk.Label(
                label="Access Key",
                halign=Gtk.Align.START,
                valign=Gtk.Align.CENTER,
                hexpand=False,
                vexpand=False,
            ),
            0,
            1,
            1,
            1,
        )
        self._access_key_entry = self.register_widget(
            Gtk.Entry(
                halign=Gtk.Align.FILL,
                valign=Gtk.Align.CENTER,
                hexpand=True,
                vexpand=False,
            ),
            "access_key",
        )
        self._grid.attach(self._access_key_entry, 1, 1, 2, 1)

        # Secret key
        self._grid.attach(
            Gtk.Label(
                label="Secret Key",
                halign=Gtk.Align.START,
                valign=Gtk.Align.CENTER,
                hexpand=False,
                vexpand=False,
            ),
            0,
            2,
            1,
            1,
        )
        self._secret_key_entry = self.register_widget(
            Gtk.Entry(
                visibility=False,
                halign=Gtk.Align.FILL,
                valign=Gtk.Align.CENTER,
                hexpand=True,
                vexpand=False,
            ),
            "secret_key",
            exportable=False,
        )
        self._grid.attach(self._secret_key_entry, 1, 2, 2, 1)

        # Bucket name
        self._grid.attach(
            Gtk.Label(
                label="Bucket Name",
                halign=Gtk.Align.START,
                valign=Gtk.Align.CENTER,
                hexpand=False,
                vexpand=False,
            ),
            0,
            3,
            1,
            1,
        )
        widget = self.register_widget(
            Gtk.Entry(
                halign=Gtk.Align.FILL,
                valign=Gtk.Align.CENTER,
                hexpand=True,
                vexpand=False,
            ),
            "bucket_name",
        )
        self._grid.attach(widget, 1, 3, 1, 1)
        widget = self.register_widget(
            Gtk.CheckButton(
                active=False,
                label="Create bucket if necessary",
                halign=Gtk.Align.START,
                valign=Gtk.Align.CENTER,
                hexpand=False,
                vexpand=False,
            ),
            "force_bucket_creation",
        )
        self._grid.attach(widget, 2, 3, 1, 1)

    @classmethod
    def _get_dict_tagset(
        cls, preflight_check_metadata: Dict[int, Dict[str, Any]], tagtype: str
    ) -> Optional[dict]:
        tags = query_metadata(preflight_check_metadata, tagtype)
        if tags is None:
            return None
        tagset = [dict(Key=_key, Value=_value) for _key, _value in tags.items()]
        return dict(TagSet=tagset)

    @classmethod
    def _get_dict_acl_options(
        cls,
        preflight_check_metadata: Dict[int, Dict[str, Any]],
        resource: str,
        allow_list: Sequence[str],
    ) -> Optional[dict]:
        options = query_metadata(preflight_check_metadata, resource)
        if options is None:
            return None
        for option in options:
            if option not in allow_list:
                raise ValueError(f"{option} is not permitted in {resource}")
        return options

    @classmethod
    def _get_dict_bucket_cors(
        cls, preflight_check_metadata: Dict[int, Dict[str, Any]]
    ) -> Optional[dict]:
        options = query_metadata(preflight_check_metadata, "bucket_cors")
        if options is None:
            return None
        if not isinstance(options, dict) or "CORSRules" not in options:
            raise ValueError(
                "bucket_cors does not contain a valid CORS configuration"
            )
        return options

    @classmethod
    def _get_client_options(cls, params: Munch) -> dict:
        client_options = dict()
        client_options["endpoint_url"] = params.hostname
        client_options["verify"] = params.hostname_ssl_verify
        client_options["aws_access_key_id"] = params.access_key
        client_options["aws_secret_access_key"] = params.secret_key
        return client_options

    @classmethod
    def _preflight_check(
        cls,
        preflight_check_metadata: Dict[int, Dict[str, Any]],
        self: S3UploaderOperation,
        params: Munch,
    ):
        # this variable will be used later on in preflight_cleanup
        self._bucket_created = False

        client_options = cls._get_client_options(params)

        # bucket creation options
        bucket_acl_options = cls._get_dict_acl_options(
            preflight_check_metadata,
            "bucket_acl_options",
            ALLOWED_BUCKET_ACL_OPTIONS,
        )
        # object creation options
        object_acl_options = cls._get_dict_acl_options(
            preflight_check_metadata,
            "object_acl_options",
            ALLOWED_OBJECT_ACL_OPTIONS,
        )

        # see https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Client.put_bucket_tagging
        bucket_tags = cls._get_dict_tagset(
            preflight_check_metadata, "bucket_tags"
        )
        # see https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Client.put_object_tagging
        object_tags = cls._get_dict_tagset(
            preflight_check_metadata, "object_tags"
        )

        # get CORS bucket configuration
        # see https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Client.put_bucket_cors
        bucket_cors = cls._get_dict_bucket_cors(preflight_check_metadata)

        # open connection
        s3_client = boto3.client("s3", **client_options)

        # check if the bucket exists
        # taken from https://stackoverflow.com/a/47565719
        try:
            s3_client.head_bucket(Bucket=params.bucket_name)
        except botocore.exceptions.ClientError as e:
            # If a client error is thrown, then check that it was a 404 error.
            # If it was a 404 error, then the bucket does not exist.
            error_code = int(e.response["Error"]["Code"])
            if error_code == 403:
                raise
            elif error_code == 404:
                if params.force_bucket_creation:
                    s3_client.create_bucket(Bucket=params.bucket_name)
                    self._bucket_created = True
                    if bucket_tags:
                        s3_client.put_bucket_tagging(
                            Bucket=params.bucket_name, Tagging=bucket_tags
                        )
                    if bucket_acl_options:
                        s3_client.put_bucket_acl(
                            Bucket=params.bucket_name, **bucket_acl_options
                        )
                    if bucket_cors:
                        s3_client.put_bucket_cors(
                            Bucket=params.bucket_name,
                            CORSConfiguration=bucket_cors,
                        )
                else:
                    raise
            else:
                raise
        # try uploading a simple object
        with tempfile.NamedTemporaryFile(delete=False) as f:
            f.write(os.urandom(1024))  # 1 kB
            tmpfile = f.name
        try:
            s3_client.upload_file(
                Filename=tmpfile,
                Bucket=params.bucket_name,
                Key=os.path.basename(tmpfile),
                Config=TransferConfig,
                ExtraArgs={
                    "Metadata": {
                        AWS_S3_ENGINE_IGNORE_ME: "1",
                    }
                },
            )
            if object_tags:
                s3_client.put_object_tagging(
                    Bucket=params.bucket_name,
                    Key=os.path.basename(tmpfile),
                    Tagging=object_tags,
                )
            if object_acl_options:
                s3_client.put_object_acl(
                    Bucket=params.bucket_name,
                    Key=os.path.basename(tmpfile),
                    **object_acl_options,
                )
        except:
            raise
        else:
            # if successful, remove it
            # delete tags first!
            if object_tags:
                s3_client.delete_object_tagging(
                    Bucket=params.bucket_name,
                    Key=os.path.basename(tmpfile),
                )

            s3_client.delete_object(
                Bucket=params.bucket_name,
                Key=os.path.basename(tmpfile),
            )
        finally:
            os.unlink(tmpfile)

    def preflight_check(self):
        self._preflight_check(
            self.appwindow.preflight_check_metadata, self, self.params
        )

    @classmethod
    def _attach_metadata(
        cls,
        file: File,
        endpoint_url: str,
        key: str,
        params: Munch,
        operation_index: int,
    ):
        parsed_url = urllib.parse.urlparse(endpoint_url)
        file.operation_metadata[operation_index] = {
            "s3 object url": f"{parsed_url.scheme}://{params.bucket_name}.{parsed_url.netloc}/{urllib.parse.quote(key)}"
        }

    @classmethod
    def _preflight_cleanup(
        cls, success: bool, self: S3UploaderOperation, params: Munch
    ):
        if not success and self._bucket_created:
            # preflight_check failed and a bucket was created: delete it!
            client_options = cls._get_client_options(params)

            # open connection
            s3_client = boto3.client("s3", **client_options)

            try:
                s3_client.delete_bucket(Bucket=params.bucket_name)
                logger.info(f"Bucket {params.bucket_name} successfully deleted")
            except Exception:
                logger.exception(f"Could not delete {params.bucket_name}")

            del s3_client
        return

    def preflight_cleanup(self, success: bool):
        self._preflight_cleanup(success, self, self.params)

    @classmethod
    @retry(
        retry=monitor_retry_condition(),
        reraise=True,
        stop=stop_after_attempt(5),
        wait=wait_exponential(),
        before=before_log(logger, logging.DEBUG),
        after=after_log(logger, logging.DEBUG),
        before_sleep=before_sleep_log(logger, logging.DEBUG),
    )
    def _run(
        cls,
        file: File,
        preflight_check_metadata: Dict[int, Dict[str, Any]],
        params: Munch,
        operation_index: int,
    ):
        client_options = cls._get_client_options(params)
        s3_client = boto3.client("s3", **client_options)

        # object creation options
        object_acl_options = cls._get_dict_acl_options(
            preflight_check_metadata,
            "object_acl_options",
            ALLOWED_OBJECT_ACL_OPTIONS,
        )
        # see https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Client.put_object_tagging
        object_tags = cls._get_dict_tagset(
            preflight_check_metadata, "object_tags"
        )

        key = str(PurePosixPath(*file.relative_filename.parts))

        # Check if file already exists and is identical
        # Inspired by: https://stackoverflow.com/questions/6591047/etag-definition-changed-in-amazon-s3
        try:
            response = s3_client.head_object(
                Bucket=params.bucket_name,
                Key=key,
            )
        except botocore.exceptions.ClientError as e:
            # key not found, which is fine
            if int(e.response["Error"]["Code"]) != 404:
                return str(e)
        else:
            if Path(file.filename).stat().st_size == int(
                response["ContentLength"]
            ):
                remote_etag = response["ETag"][
                    1:-1
                ]  # get rid of those extra quotes
                local_etag = calculate_etag(file.filename)
                if remote_etag == local_etag:
                    # attach metadata
                    cls._attach_metadata(
                        file,
                        client_options["endpoint_url"],
                        key,
                        params,
                        operation_index,
                    )
                    raise SkippedOperation("File has been uploaded already")

        try:
            s3_client.upload_file(
                Filename=file.filename,
                Bucket=params.bucket_name,
                Key=key,
                Config=TransferConfig,
                Callback=S3ProgressPercentage(
                    file, file.filename, operation_index
                ),
            )
            if object_tags:
                s3_client.put_object_tagging(
                    Bucket=params.bucket_name,
                    Key=key,
                    Tagging=object_tags,
                )
            if object_acl_options:
                s3_client.put_object_acl(
                    Bucket=params.bucket_name,
                    Key=key,
                    **object_acl_options,
                )
        except Exception as e:
            logger.exception(f"S3UploaderOperation.run exception")
            del s3_client
            del client_options
            return str(e)
        else:
            # add object URL to metadata
            cls._attach_metadata(
                file,
                client_options["endpoint_url"],
                key,
                params,
                operation_index,
            )
            del s3_client
            del client_options
        return None

    @add_directory_support
    def run(self, file: File):
        return self._run(
            file,
            self.appwindow.preflight_check_metadata,
            self.params,
            self.index,
        )



import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk

from ..engine_advanced_settings import EngineAdvancedSettings
from ..engine import Engine
from ..utils import PATTERN_PLACEHOLDER_TEXT


class DirectoryWatchdogEngineAdvancedSettings(EngineAdvancedSettings):
    def __init__(self, engine: Engine):
        super().__init__(engine)

        self._row_counter = 0

        # Process existing directories in monitored directory
        self._process_existing_directories_checkbutton = engine.register_widget(
            Gtk.CheckButton(
                label="Process existing directories in target directory",
                halign=Gtk.Align.FILL,
                valign=Gtk.Align.CENTER,
                hexpand=True,
                vexpand=False,
                active=False,
            ),
            "process_existing_directories",
        )
        self.attach(
            self._process_existing_directories_checkbutton,
            0,
            self._row_counter,
            1,
            1,
        )
        self._row_counter += 1

        self._add_horizontal_separator()

        # Specify allowed file patterns
        allowed_file_patterns_grid = Gtk.Grid(
            halign=Gtk.Align.FILL,
            valign=Gtk.Align.CENTER,
            hexpand=True,
            vexpand=False,
            column_spacing=5,
        )

        self.attach(allowed_file_patterns_grid, 0, self._row_counter, 1, 1)
        self._row_counter += 1
        allowed_file_patterns_grid.attach(
            Gtk.Label(
                label="Allowed filename patterns",
                halign=Gtk.Align.START,
                valign=Gtk.Align.CENTER,
                hexpand=False,
                vexpand=False,
            ),
            0,
            0,
            1,
            1,
        )
        self._allowed_file_patterns_entry = engine.register_widget(
            Gtk.Entry(
                placeholder_text=PATTERN_PLACEHOLDER_TEXT,
                halign=Gtk.Align.FILL,
                valign=Gtk.Align.CENTER,
                hexpand=True,
                vexpand=False,
            ),
            "allowed_file_patterns",
        )
        allowed_file_patterns_grid.attach(
            self._allowed_file_patterns_entry, 1, 0, 1, 1
        )

        self._add_horizontal_separator()

        # Specify ignored file patterns
        ignore_file_patterns_grid = Gtk.Grid(
            halign=Gtk.Align.FILL,
            valign=Gtk.Align.CENTER,
            hexpand=True,
            vexpand=False,
            column_spacing=5,
        )

        self.attach(ignore_file_patterns_grid, 0, self._row_counter, 1, 1)
        self._row_counter += 1
        ignore_file_patterns_grid.attach(
            Gtk.Label(
                label="Ignored filename patterns",
                halign=Gtk.Align.START,
                valign=Gtk.Align.CENTER,
                hexpand=False,
                vexpand=False,
            ),
            0,
            0,
            1,
            1,
        )
        self._ignored_file_patterns_entry = engine.register_widget(
            Gtk.Entry(
                placeholder_text=PATTERN_PLACEHOLDER_TEXT,
                halign=Gtk.Align.FILL,
                valign=Gtk.Align.CENTER,
                hexpand=True,
                vexpand=False,
            ),
            "ignore_file_patterns",
        )
        ignore_file_patterns_grid.attach(
            self._ignored_file_patterns_entry, 1, 0, 1, 1
        )

        self._add_horizontal_separator()

        # Specify allowed directory patterns
        allowed_directory_patterns_grid = Gtk.Grid(
            halign=Gtk.Align.FILL,
            valign=Gtk.Align.CENTER,
            hexpand=True,
            vexpand=False,
            column_spacing=5,
        )

        self.attach(allowed_directory_patterns_grid, 0, self._row_counter, 1, 1)
        self._row_counter += 1
        allowed_directory_patterns_grid.attach(
            Gtk.Label(
                label="Allowed directory patterns",
                halign=Gtk.Align.START,
                valign=Gtk.Align.CENTER,
                hexpand=False,
                vexpand=False,
            ),
            0,
            0,
            1,
            1,
        )
        self._allowed_directory_patterns_entry = engine.register_widget(
            Gtk.Entry(
                placeholder_text=PATTERN_PLACEHOLDER_TEXT,
                halign=Gtk.Align.FILL,
                valign=Gtk.Align.CENTER,
                hexpand=True,
                vexpand=False,
            ),
            "allowed_directory_patterns",
        )
        allowed_directory_patterns_grid.attach(
            self._allowed_directory_patterns_entry, 1, 0, 1, 1
        )

        self._add_horizontal_separator()

        # Specify ignored directory patterns
        ignore_directory_patterns_grid = Gtk.Grid(
            halign=Gtk.Align.FILL,
            valign=Gtk.Align.CENTER,
            hexpand=True,
            vexpand=False,
            column_spacing=5,
        )

        self.attach(ignore_directory_patterns_grid, 0, self._row_counter, 1, 1)
        self._row_counter += 1
        ignore_directory_patterns_grid.attach(
            Gtk.Label(
                label="Ignored directory patterns",
                halign=Gtk.Align.START,
                valign=Gtk.Align.CENTER,
                hexpand=False,
                vexpand=False,
            ),
            0,
            0,
            1,
            1,
        )
        self._ignored_directory_patterns_entry = engine.register_widget(
            Gtk.Entry(
                placeholder_text=PATTERN_PLACEHOLDER_TEXT,
                halign=Gtk.Align.FILL,
                valign=Gtk.Align.CENTER,
                hexpand=True,
                vexpand=False,
            ),
            "ignore_directory_patterns",
        )
        ignore_directory_patterns_grid.attach(
            self._ignored_directory_patterns_entry, 1, 0, 1, 1
        )

    def _add_horizontal_separator(self):
        self.attach(
            Gtk.Separator(
                orientation=Gtk.Orientation.HORIZONTAL,
                halign=Gtk.Align.FILL,
                valign=Gtk.Align.CENTER,
                hexpand=True,
                vexpand=True,
            ),
            0,
            self._row_counter,
            1,
            1,
        )
        self._row_counter += 1



from __future__ import annotations

from gi.repository import GLib
import boto3
import botocore

from ..engine import Engine, EngineThread
from ..file import S3Object, FileStatus
from ..operations.s3_uploader import AWS_S3_ENGINE_IGNORE_ME
from ..utils import LongTaskWindow, match_path

import logging
from pathlib import PurePosixPath
import urllib.parse

logger = logging.getLogger(__name__)

AVAILABLE_CONFIGURATIONS = (
    "LambdaFunctionConfigurations",
    "TopicConfigurations",
    "QueueConfigurations",
)


class BaseS3BucketEngineThread(EngineThread):
    def __init__(self, engine: BaseS3BucketEngine, task_window: LongTaskWindow):
        super().__init__(engine, task_window)
        self._client_options: dict = {}

        # prepare patterns
        app = engine.appwindow.props.application
        self._included_patterns = app.get_allowed_file_patterns(
            self.params.allowed_patterns
        )
        self._excluded_patterns = app.get_ignored_file_patterns(
            self.params.ignore_patterns
        )

    def get_full_name(self, key) -> str:
        raise NotImplementedError

    def get_region_name(self, default_region: str) -> bool:
        # set up temporary s3 client
        temp_s3_client = boto3.client("s3", **self._client_options)

        # first confirm that the bucket exists and that we can read it
        try:
            temp_s3_client.head_bucket(Bucket=self.params.bucket_name)
        except Exception as e:
            self._engine.cleanup()
            GLib.idle_add(
                self._engine.abort,
                self._task_window,
                e,
                priority=GLib.PRIORITY_HIGH,
            )
            return False

        # next try to get the region the bucket is located in
        # see https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Client.get_bucket_location
        try:
            response = temp_s3_client.get_bucket_location(
                Bucket=self.params.bucket_name
            )
        except Exception as e:
            self._engine.cleanup()
            GLib.idle_add(
                self._engine.abort,
                self._task_window,
                e,
                priority=GLib.PRIORITY_HIGH,
            )
            return False

        self._client_options["region_name"] = (
            response["LocationConstraint"]
            if response["LocationConstraint"]
            else default_region
        )

        return True

    def process_new_file(self, s3_info: dict):
        object_info = s3_info["object"]
        key = urllib.parse.unquote_plus(object_info["key"])
        if "eTag" in object_info:  # AWS
            etag = object_info["eTag"]
        elif "etag" in object_info:  # Ceph
            etag = object_info["etag"]
        else:
            logger.warning(f"no etag found for {key}")
            etag = ""
        size = object_info["size"]

        if not match_path(
            PurePosixPath(key),
            included_patterns=self._included_patterns,
            excluded_patterns=self._excluded_patterns,
            case_sensitive=False,
        ):
            return False

        # ensure that this is not an S3Uploader testfile!
        try:
            response = self._engine.s3_client.head_object(
                Bucket=self.params.bucket_name,
                Key=key,
            )
        except botocore.exceptions.ClientError as e:
            error_code = int(e.response["Error"]["Code"])
            if error_code == 404:
                logger.warning(
                    f"{key} does not exist in {self.params.bucket_name}"
                )
            else:
                logger.exception(
                    f"head_object failure for {key} in {self.params.bucket_name}"
                )
            return False

        if (
            "Metadata" in response
            and AWS_S3_ENGINE_IGNORE_ME in response["Metadata"]
        ):
            # this is a testfile -> ignore!
            return False

        full_path = self.get_full_name(object_info["key"])
        relative_path = PurePosixPath(key)
        created = response["LastModified"].timestamp()

        if (
            self._engine.props.running
            and self._engine._appwindow._queue_manager.props.running
        ):
            _file = S3Object(
                full_path,
                relative_path,
                created,
                FileStatus.SAVED,
                self.params.bucket_name,
                etag,
                size,
                self._client_options["region_name"],
            )
            GLib.idle_add(
                self._engine._appwindow._queue_manager.add,
                _file,
                priority=GLib.PRIORITY_HIGH,
            )
        return True

    def process_existing_files(self) -> bool:
        GLib.idle_add(
            self._task_window.set_text, "<b>Processing existing objects...</b>"
        )
        try:
            paginator = self._engine.s3_client.get_paginator("list_objects_v2")
            page_iterator = paginator.paginate(Bucket=self.params.bucket_name)

            existing_files = []

            for page in page_iterator:
                if page["KeyCount"] == 0:
                    continue
                for _object in page["Contents"]:
                    key = _object["Key"]

                    if not match_path(
                        PurePosixPath(key),
                        included_patterns=self._included_patterns,
                        excluded_patterns=self._excluded_patterns,
                        case_sensitive=False,
                    ):
                        continue

                    last_modified = _object["LastModified"]
                    size = _object["Size"]
                    etag = _object["ETag"][
                        1:-1
                    ]  # get rid of those weird quotes
                    quoted_key = urllib.parse.quote_plus(key)

                    full_path = self.get_full_name(quoted_key)
                    relative_path = PurePosixPath(key)
                    created = last_modified.timestamp()

                    _file = S3Object(
                        full_path,
                        relative_path,
                        created,
                        FileStatus.SAVED,
                        self.params.bucket_name,
                        etag,
                        size,
                        self._client_options["region_name"],
                    )

                    existing_files.append(_file)
            if existing_files:
                GLib.idle_add(
                    self._engine._appwindow._queue_manager.add,
                    existing_files,
                    priority=GLib.PRIORITY_HIGH,
                )
        except Exception as e:
            self._engine.cleanup()
            GLib.idle_add(
                self._engine.abort,
                self._task_window,
                e,
                priority=GLib.PRIORITY_HIGH,
            )
            return False

        return True


class BaseS3BucketEngine(Engine):
    def _get_client_options(self) -> dict:
        client_options = dict()
        client_options["aws_access_key_id"] = self._get_params().access_key
        client_options["aws_secret_access_key"] = self._get_params().secret_key
        return client_options



import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, GLib
import dropbox
import keyring

from ..operation import Operation
from ..utils.exceptions import SkippedOperation
from ..queue_manager import QueueManager
from ..utils.decorators import (
    with_pango_docs,
    supported_filetypes,
    add_directory_support,
)
from ..utils import get_random_string
from ..file import File, RegularFile, Directory

import logging
import os
from pathlib import PurePosixPath
import hashlib
import re
import webbrowser
from threading import Thread

logger = logging.getLogger(__name__)

APP_KEY = "l9wed9kx1tq965c"

# taken from https://github.com/dropbox/dropbox-api-content-hasher/blob/master/python/dropbox_content_hasher.py
class DropboxContentHasher(object):
    """
    Computes a hash using the same algorithm that the Dropbox API uses for the
    the "content_hash" metadata field.
    The digest() method returns a raw binary representation of the hash.  The
    hexdigest() convenience method returns a hexadecimal-encoded version, which
    is what the "content_hash" metadata field uses.
    This class has the same interface as the hashers in the standard 'hashlib'
    package.
    Example:
        hasher = DropboxContentHasher()
        with open('some-file', 'rb') as f:
            while True:
                chunk = f.read(1024)  # or whatever chunk size you want
                if len(chunk) == 0:
                    break
                hasher.update(chunk)
        print(hasher.hexdigest())
    """

    BLOCK_SIZE = 4 * 1024 * 1024

    def __init__(self):
        self._overall_hasher = hashlib.sha256()
        self._block_hasher = hashlib.sha256()
        self._block_pos = 0

        self.digest_size = self._overall_hasher.digest_size

    def update(self, new_data):
        if self._overall_hasher is None:
            raise AssertionError(
                "can't use this object anymore; you already called digest()"
            )

        new_data_pos = 0
        while new_data_pos < len(new_data):
            if self._block_pos == self.BLOCK_SIZE:
                self._overall_hasher.update(self._block_hasher.digest())
                self._block_hasher = hashlib.sha256()
                self._block_pos = 0

            space_in_block = self.BLOCK_SIZE - self._block_pos
            part = new_data[new_data_pos : (new_data_pos + space_in_block)]
            self._block_hasher.update(part)

            self._block_pos += len(part)
            new_data_pos += len(part)

    def _finish(self):
        if self._overall_hasher is None:
            raise AssertionError(
                "can't use this object anymore; you already called digest() or hexdigest()"
            )

        if self._block_pos > 0:
            self._overall_hasher.update(self._block_hasher.digest())
            self._block_hasher = None
        h = self._overall_hasher
        self._overall_hasher = (
            None  # Make sure we can't use this object anymore.
        )
        return h

    def digest(self):
        return self._finish().digest()

    def hexdigest(self):
        return self._finish().hexdigest()


class DropboxLinkDialog(Gtk.Dialog):
    def __init__(self, appwindow):
        Gtk.Dialog.__init__(
            self,
            title="Link Dropbox",
            transient_for=appwindow,
            modal=True,
            flags=0,
        )
        self._authorization_code = None
        self._okbutton = self.add_button("Confirm", Gtk.ResponseType.OK)
        self._okbutton.set_sensitive(False)

        self.set_default_size(150, 100)

        box = self.get_content_area()
        grid = Gtk.Grid(
            row_spacing=5,
            column_spacing=5,
            border_width=5,
            hexpand=True,
            vexpand=True,
            halign=Gtk.Align.FILL,
            valign=Gtk.Align.FILL,
        )
        box.add(grid)

        label = Gtk.Label(
            label="<b>Follow these steps to link your Dropbox account with the RFI-File-Monitor</b>",
            use_markup=True,
            halign=Gtk.Align.FILL,
            valign=Gtk.Align.CENTER,
            hexpand=False,
            vexpand=False,
        )
        frame = Gtk.Frame(
            height_request=50,
            border_width=5,
            halign=Gtk.Align.FILL,
            valign=Gtk.Align.START,
            hexpand=True,
            vexpand=False,
        )
        frame.add(label)
        grid.attach(frame, 0, 0, 1, 1)

        grid1 = Gtk.Grid(
            row_spacing=5,
            column_spacing=5,
            hexpand=True,
            vexpand=False,
            halign=Gtk.Align.START,
            valign=Gtk.Align.CENTER,
        )
        label = Gtk.Label(
            label="1. Click the button to open Dropbox in your browser",
            halign=Gtk.Align.FILL,
            valign=Gtk.Align.CENTER,
            hexpand=True,
            vexpand=False,
        )
        grid1.attach(label, 0, 0, 1, 1)
        dbx_button = Gtk.Button(
            label="Open Dropbox",
            halign=Gtk.Align.END,
            valign=Gtk.Align.CENTER,
            hexpand=False,
            vexpand=False,
        )
        grid1.attach(dbx_button, 1, 0, 1, 1)
        grid.attach(grid1, 0, 1, 1, 1)
        dbx_button.connect("clicked", self._dropbox_button_clicked)

        label = Gtk.Label(
            label='2. Click "Allow" (you may have to log in first).',
            halign=Gtk.Align.START,
            valign=Gtk.Align.CENTER,
            hexpand=True,
            vexpand=False,
        )
        grid.attach(label, 0, 2, 1, 1)

        grid3 = Gtk.Grid(
            row_spacing=5,
            column_spacing=5,
            hexpand=True,
            vexpand=False,
            halign=Gtk.Align.FILL,
            valign=Gtk.Align.CENTER,
        )
        label = Gtk.Label(
            label="3. Copy the authorization token into the box",
            halign=Gtk.Align.START,
            valign=Gtk.Align.CENTER,
            hexpand=False,
            vexpand=False,
        )
        grid3.attach(label, 0, 0, 1, 1)
        self._entry = Gtk.Entry(
            halign=Gtk.Align.FILL,
            valign=Gtk.Align.CENTER,
            hexpand=True,
            vexpand=False,
        )
        self._entry.set_sensitive(False)
        self._entry.connect("changed", self._dropbox_entry_changed)
        grid3.attach(self._entry, 1, 0, 1, 1)
        grid.attach(grid3, 0, 3, 1, 1)

        self.show_all()

    def _dropbox_entry_changed(self, entry):
        self._okbutton.set_sensitive(True)
        self._authorization_code = entry.get_text().strip()

    def _dropbox_button_clicked(self, button):
        self._auth_flow = dropbox.oauth.DropboxOAuth2FlowNoRedirect(
            consumer_key=APP_KEY,
            use_pkce=True,
            token_access_type="offline",
        )
        url = self._auth_flow.start()
        webbrowser.open_new_tab(url)
        button.set_sensitive(False)
        self._entry.set_sensitive(True)

    @property
    def authorization_code(self) -> str:
        return self._authorization_code

    @property
    def auth_flow(self) -> dropbox.oauth.DropboxOAuth2FlowNoRedirect:
        return self._auth_flow


# taken from Maestral
class DropboxSpaceUsage(dropbox.users.SpaceUsage):
    @property
    def allocation_type(self) -> str:
        if self.allocation.is_team():
            return "team"
        elif self.allocation.is_individual():
            return "individual"
        else:
            return ""

    def __str__(self) -> str:

        if self.allocation.is_individual():
            used = self.used
            allocated = self.allocation.get_individual().allocated
        elif self.allocation.is_team():
            used = self.allocation.get_team().used
            allocated = self.allocation.get_team().allocated
        else:
            return self.natural_size(self.used)

        percent = used / allocated
        if percent <= 0.8:
            color = "green"
        elif percent <= 0.95:
            color = "orange"
        else:
            color = "red"
        return f'<span weight="bold" foreground="{color}">{percent:.1%} of {self.natural_size(allocated)}</span>'

    @classmethod
    def natural_size(cls, num: float, unit: str = "B", sep: bool = True) -> str:
        """
        Convert number to a human readable string with decimal prefix.
        :param float num: Value in given unit.
        :param unit: Unit suffix.
        :param sep: Whether to separate unit and value with a space.
        :returns: Human readable string with decimal prefixes.
        """
        sep_char = " " if sep else ""

        for prefix in ("", "K", "M", "G"):
            if abs(num) < 1000.0:
                return f"{num:3.1f}{sep_char}{prefix}{unit}"
            num /= 1000.0

        prefix = "T"
        return f"{num:.1f}{sep_char}{prefix}{unit}"

    @classmethod
    def from_dbx_space_usage(cls, su: dropbox.users.SpaceUsage):
        return cls(used=su.used, allocation=su.allocation)


class DropboxSpaceCheckerThread(Thread):
    def __init__(self, operation):
        super().__init__()
        self._operation = operation

    def run(self):
        if self._operation._dropbox:
            try:
                usage = DropboxSpaceUsage.from_dbx_space_usage(
                    self._operation._dropbox.users_get_space_usage()
                )
            except Exception as e:
                usage = f"error getting usage: {str(e)}"
        else:
            usage = "not available"
        GLib.idle_add(
            self._operation._update_space_usage,
            usage,
            priority=GLib.PRIORITY_DEFAULT_IDLE,
        )


@with_pango_docs(filename="dropbox_uploader.pango")
@supported_filetypes(filetypes=[RegularFile, Directory])
class DropboxUploaderOperation(Operation):
    NAME = "Dropbox Uploader"

    SESSION = dropbox.create_session(max_connections=QueueManager.MAX_JOBS)
    CHUNK_SIZE = 1024 * 1024  # 1MB

    def __init__(self, *args, **kwargs):
        Operation.__init__(self, *args, **kwargs)

        self._dropbox = None
        self._space_thread = None

        self._grid = Gtk.Grid(
            border_width=5,
            row_spacing=5,
            column_spacing=5,
            halign=Gtk.Align.FILL,
            valign=Gtk.Align.CENTER,
            hexpand=True,
            vexpand=False,
        )
        self.add(self._grid)

        # In the initial version, use only access-key and folder name
        self._grid.attach(
            Gtk.Label(
                label="Destination folder",
                halign=Gtk.Align.START,
                valign=Gtk.Align.CENTER,
                hexpand=False,
                vexpand=False,
            ),
            0,
            0,
            1,
            1,
        )
        widget = self.register_widget(
            Gtk.Entry(
                halign=Gtk.Align.FILL,
                valign=Gtk.Align.CENTER,
                hexpand=True,
                vexpand=False,
            ),
            "destination_folder",
        )
        self._grid.attach(widget, 1, 0, 2, 1)

        self._grid.attach(
            Gtk.Label(
                label="Email address",
                halign=Gtk.Align.START,
                valign=Gtk.Align.CENTER,
                hexpand=False,
                vexpand=False,
            ),
            0,
            1,
            1,
            1,
        )
        self._email_entry = self.register_widget(
            Gtk.Entry(
                placeholder_text="Address used for registering with Dropbox",
                halign=Gtk.Align.FILL,
                valign=Gtk.Align.CENTER,
                hexpand=True,
                vexpand=False,
            ),
            "email",
        )
        self._grid.attach(self._email_entry, 1, 1, 1, 1)
        self._email_entry.connect("changed", self._email_entry_changed_cb)

        button = Gtk.Button(
            label="Validate",
            halign=Gtk.Align.END,
            valign=Gtk.Align.CENTER,
            hexpand=False,
            vexpand=False,
        )
        self._grid.attach(button, 2, 1, 1, 1)
        button.connect("clicked", self._validate_button_clicked_cb)

        self._space_label = Gtk.Label(
            label="Space Usage: not available",
            use_markup=True,
            halign=Gtk.Align.START,
            valign=Gtk.Align.CENTER,
            hexpand=True,
            vexpand=True,
        )
        self._grid.attach(self._space_label, 0, 2, 3, 1)

        # check space usage every 60 seconds and update the corresponding label
        GLib.timeout_add_seconds(
            60,
            self._launch_space_usage_thread,
            False,
            priority=GLib.PRIORITY_DEFAULT_IDLE,
        )

    def _launch_space_usage_thread(self, kill):
        space_thread = DropboxSpaceCheckerThread(self)
        space_thread.start()

        if kill:
            return GLib.SOURCE_REMOVE
        else:
            return GLib.SOURCE_CONTINUE

    def _update_space_usage(self, usage):
        self._space_label.props.label = f"Space Usage: {str(usage)}"

        return GLib.SOURCE_REMOVE

    def _email_entry_changed_cb(self, entry):
        # any changes to the email address reset the Dropbox client
        self._dropbox = None
        self._email_entry.set_icon_from_icon_name(
            icon_pos=Gtk.EntryIconPosition.SECONDARY,
            icon_name="emblem-unreadable",
        )

    def _validate_button_clicked_cb(self, button):
        # first confirm that we have a proper email address
        if not re.fullmatch(r"[^@]+@[^@]+\.[^@]+", self.params.email):
            dialog = Gtk.MessageDialog(
                transient_for=self.appwindow,
                modal=True,
                destroy_with_parent=True,
                message_type=Gtk.MessageType.ERROR,
                buttons=Gtk.ButtonsType.CLOSE,
                text=f"{self.params.email} is not a vaild email address",
            )
            dialog.run()
            dialog.destroy()
            self._email_entry.set_icon_from_icon_name(
                icon_pos=Gtk.EntryIconPosition.SECONDARY,
                icon_name="emblem-unreadable",
            )
            return

        # check keyring
        try:
            refresh_token = keyring.get_password(
                "RFI-File-Monitor-Dropbox", self.params.email.lower()
            )
        except keyring.errors.KeyringError as e:
            dialog = Gtk.MessageDialog(
                transient_for=self.appwindow,
                modal=True,
                destroy_with_parent=True,
                message_type=Gtk.MessageType.ERROR,
                buttons=Gtk.ButtonsType.CLOSE,
                text="Error accessing keyring",
                secondary_text=str(e),
            )
            dialog.run()
            dialog.destroy()
            self._email_entry.set_icon_from_icon_name(
                icon_pos=Gtk.EntryIconPosition.SECONDARY,
                icon_name="emblem-unreadable",
            )
            return

        if refresh_token:
            # use refresh token to launch dropbox session
            self._dropbox = dropbox.Dropbox(
                oauth2_refresh_token=refresh_token,
                session=self.SESSION,
                app_key=APP_KEY,
            )
            exc = None
            try:
                account_info = self._dropbox.users_get_current_account()
            except dropbox.exceptions.AuthError as e:
                if e.error.is_invalid_access_token():
                    logger.info("Dropbox token has been revoked!")
                    self._dropbox = None
                    # delete keyring password
                    keyring.delete_password(
                        "RFI-File-Monitor-Dropbox", self.params.email.lower()
                    )
                    self._validate_button_clicked_cb(button)
                    return
                exc = e
            except Exception as e:
                exc = e
            if exc:
                dialog = Gtk.MessageDialog(
                    transient_for=self.appwindow,
                    modal=True,
                    destroy_with_parent=True,
                    message_type=Gtk.MessageType.ERROR,
                    buttons=Gtk.ButtonsType.CLOSE,
                    text="Could not get user information",
                    secondary_text=str(exc),
                )
                dialog.run()
                dialog.destroy()
                self._email_entry.set_icon_from_icon_name(
                    icon_pos=Gtk.EntryIconPosition.SECONDARY,
                    icon_name="emblem-unreadable",
                )
                self._dropbox = None
                return

            if account_info.email.lower() != self.params.email.lower():
                dialog = Gtk.MessageDialog(
                    transient_for=self.appwindow,
                    modal=True,
                    destroy_with_parent=True,
                    message_type=Gtk.MessageType.ERROR,
                    buttons=Gtk.ButtonsType.CLOSE,
                    text="Email address does not match Dropbox user records",
                )
                dialog.run()
                dialog.destroy()
                self._email_entry.set_icon_from_icon_name(
                    icon_pos=Gtk.EntryIconPosition.SECONDARY,
                    icon_name="emblem-unreadable",
                )
                self._dropbox = None
            else:
                self._email_entry.set_icon_from_icon_name(
                    icon_pos=Gtk.EntryIconPosition.SECONDARY,
                    icon_name="emblem-default",
                )
                GLib.idle_add(
                    self._launch_space_usage_thread,
                    True,
                    priority=GLib.PRIORITY_DEFAULT_IDLE,
                )
            return

        # without refresh token -> link app
        dbx_dialog = DropboxLinkDialog(self.appwindow)
        if dbx_dialog.run() != Gtk.ResponseType.OK:
            # not OK means that the auth flow was not started or aborted
            self._email_entry.set_icon_from_icon_name(
                icon_pos=Gtk.EntryIconPosition.SECONDARY,
                icon_name="emblem-unreadable",
            )
            self._dropbox = None
            dbx_dialog.destroy()
            return

        authorization_code = dbx_dialog.authorization_code
        auth_flow = dbx_dialog.auth_flow
        dbx_dialog.destroy()
        try:
            res = auth_flow.finish(authorization_code)
        except Exception as e:
            dialog = Gtk.MessageDialog(
                transient_for=self.appwindow,
                modal=True,
                destroy_with_parent=True,
                message_type=Gtk.MessageType.ERROR,
                buttons=Gtk.ButtonsType.CLOSE,
                text="Could not link RFI-File-Monitor to Dropbox",
                secondary_text=str(e),
            )
            dialog.run()
            dialog.destroy()
            self._email_entry.set_icon_from_icon_name(
                icon_pos=Gtk.EntryIconPosition.SECONDARY,
                icon_name="emblem-unreadable",
            )
            self._dropbox = None
            return

        refresh_token = res.refresh_token
        self._dropbox = dropbox.Dropbox(
            oauth2_refresh_token=refresh_token,
            session=self.SESSION,
            app_key=APP_KEY,
        )
        self._email_entry.set_icon_from_icon_name(
            icon_pos=Gtk.EntryIconPosition.SECONDARY, icon_name="emblem-default"
        )
        GLib.idle_add(
            self._launch_space_usage_thread,
            True,
            priority=GLib.PRIORITY_DEFAULT_IDLE,
        )
        # save token in keyring
        try:
            keyring.set_password(
                "RFI-File-Monitor-Dropbox",
                self.params.email.lower(),
                refresh_token,
            )
        except keyring.errors.KeyringError as e:
            dialog = Gtk.MessageDialog(
                transient_for=self.appwindow,
                modal=True,
                destroy_with_parent=True,
                message_type=Gtk.MessageType.ERROR,
                buttons=Gtk.ButtonsType.CLOSE,
                text="Error accessing keyring. Dropbox should still work in this session though.",
                secondary_text=str(e),
            )
            dialog.run()
            dialog.destroy()

    def preflight_check(self):
        if not self.params.destination_folder:
            raise Exception("Destination folder cannot be an empty string")

        if not self._dropbox:
            raise Exception(
                "Validate the email address and link the Dropbox account"
            )

        self._dropbox.check_user("test-user")

        # create folder and upload test file
        self._base_folder = f"/{self.params.destination_folder}"
        try:
            self._dropbox.files_create_folder_v2(self._base_folder)
        except dropbox.exceptions.ApiError:
            # an exception is thrown if the directory already exists
            pass
        test_filename = (
            f"{self._base_folder}/test-file-{get_random_string(6)}.txt"
        )
        self._dropbox.files_upload(b"Dummy contents", path=test_filename)

        # delete file
        self._dropbox.files_delete_v2(test_filename)

    @add_directory_support
    def run(self, file: File):
        # check first if file already exists
        # if it does, then first get its size
        # if the size matches, calculate checksum.
        # if checksums match, SKIP
        # else: remove file and upload new version

        dbx_filename = str(
            PurePosixPath(self._base_folder, *file.relative_filename.parts)
        )
        size = os.path.getsize(file.filename)

        try:
            metadata = self._dropbox.files_get_metadata(dbx_filename)
        except dropbox.exceptions.ApiError:
            pass
        else:
            if size == metadata.size:
                # calculate content hash of local file
                hasher = DropboxContentHasher()
                with open(file.filename, "rb") as f:
                    while True:
                        chunk = f.read(4096)
                        if len(chunk) == 0:
                            break
                        hasher.update(chunk)
                if hasher.hexdigest() == metadata.content_hash:
                    raise SkippedOperation(
                        "File has already been uploaded to Dropbox"
                    )
                else:
                    # delete remote file
                    self._dropbox.files_delete_v2(dbx_filename)
            else:
                # delete remote file
                self._dropbox.files_delete_v2(dbx_filename)

        # the following was inspired by Maestral
        if size <= self.CHUNK_SIZE:
            # upload small file
            with open(file.filename, "rb") as f:
                self._dropbox.files_upload(f.read(), dbx_filename)
        else:
            # upload large file
            with open(file.filename, "rb") as f:
                session_start = self._dropbox.files_upload_session_start(
                    f.read(self.CHUNK_SIZE)
                )
                uploaded = f.tell()

                cursor = dropbox.files.UploadSessionCursor(
                    session_id=session_start.session_id, offset=uploaded
                )
                commit = dropbox.files.CommitInfo(path=dbx_filename)

                while True:
                    try:
                        if size - f.tell() <= self.CHUNK_SIZE:
                            md = self._dropbox.files_upload_session_finish(
                                f.read(self.CHUNK_SIZE), cursor, commit
                            )

                        else:
                            self._dropbox.files_upload_session_append_v2(
                                f.read(self.CHUNK_SIZE), cursor
                            )
                            md = None

                        # housekeeping
                        uploaded = f.tell()

                        # upload progressbar
                        file.update_progressbar(
                            self.index, 100 * uploaded / size
                        )

                        if md:
                            break
                        else:
                            cursor.offset = uploaded

                    except dropbox.exceptions.DropboxException as exc:
                        error = getattr(exc, "error", None)
                        if (
                            isinstance(
                                error, dropbox.files.UploadSessionFinishError
                            )
                            and error.is_lookup_failed()
                        ):
                            session_lookup_error = error.get_lookup_failed()
                        elif isinstance(
                            error, dropbox.files.UploadSessionLookupError
                        ):
                            session_lookup_error = error
                        else:
                            return str(exc)

                        if session_lookup_error.is_incorrect_offset():
                            o = (
                                session_lookup_error.get_incorrect_offset().correct_offset
                            )
                            # reset position in file
                            f.seek(o)
                            cursor.offset = f.tell()
                        else:
                            return str(exc)

        return None



import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, GLib, Gio
from .utils import match_path

from enum import auto, IntEnum, unique
import logging
from pathlib import PurePath, PurePosixPath, Path
from typing import Final, Dict, Any, Optional, List, Tuple
from abc import ABC, abstractmethod
from time import time


logger = logging.getLogger(__name__)


@unique
class FileStatus(IntEnum):
    CREATED = auto()
    SAVED = auto()
    QUEUED = auto()
    RUNNING = auto()
    SUCCESS = auto()
    FAILURE = auto()
    REMOVED_FROM_LIST = auto()
    SKIPPED = auto()

    def __str__(self):
        # pylint: disable=no-member
        return self.name.lower().capitalize().replace("_", " ")


class File(ABC):
    @abstractmethod
    def __init__(
        self,
        filename: str,
        relative_filename: PurePath,
        created: float,
        status: FileStatus,
    ):

        self._filename = filename
        self._relative_filename = relative_filename
        self._created: Final[float] = created
        self._status = status
        self._row_reference: Gtk.TreeRowReference = None
        self._operation_metadata: Final[Dict[int, Dict[str, Any]]] = dict()
        self._cancellable = Gio.Cancellable()
        self._saved: float = 0
        self._requeue: bool = False
        self._succeeded: float = 0

    @property
    def cancellable(self) -> Gio.Cancellable:
        return self._cancellable

    @property
    def operation_metadata(self) -> Dict[int, Dict[str, Any]]:
        return self._operation_metadata

    @property
    def filename(self) -> str:
        return self._filename

    @property
    def relative_filename(self) -> PurePath:
        return self._relative_filename

    @property
    def created(self) -> float:
        return self._created

    @property
    def saved(self) -> float:
        return self._saved

    @saved.setter
    def saved(self, value: int):
        self._saved = value

    @property
    def succeeded(self) -> float:
        return self._succeeded

    @succeeded.setter
    def succeeded(self, value: int):
        self._succeeded = value

    @property
    def status(self) -> FileStatus:
        return self._status

    @status.setter
    def status(self, value: FileStatus):
        self._status = value

    @property
    def requeue(self) -> bool:
        return self._requeue

    @requeue.setter
    def requeue(self, value: bool):
        self._requeue = value

    @property
    def row_reference(self) -> Gtk.TreeRowReference:
        return self._row_reference

    @row_reference.setter
    def row_reference(self, value: Gtk.TreeRowReference):
        self._row_reference = value

    def __str__(self):
        return f"{type(self).__name__}: {self._filename} -> {str(self._status)}"

    def _update_progressbar_worker_cb(self, index: int, value: float):
        if not self.row_reference.valid():
            logger.warning(
                f"_update_progressbar_worker_cb: {self.filename} is invalid!"
            )
            return GLib.SOURCE_REMOVE

        model = self.row_reference.get_model()
        path = self.row_reference.get_path()
        parent_iter = model.get_iter(path)
        n_children = model.iter_n_children(parent_iter)

        cumul_value = (index * 100.0 + value) / n_children
        model[parent_iter][4] = cumul_value
        model[parent_iter][5] = f"{cumul_value:.1f} %"

        child_iter = model.iter_nth_child(parent_iter, index)
        model[child_iter][4] = value
        model[child_iter][5] = f"{value:.1f} %"

        return GLib.SOURCE_REMOVE

    def _update_status_worker_cb(self, index: int, status: FileStatus, message):
        if not self.row_reference.valid():
            logger.warning(
                f"_update_status_worker_cb: {self.filename} is invalid!"
            )
            return GLib.SOURCE_REMOVE

        model = self.row_reference.get_model()
        path = self.row_reference.get_path()
        iter = model.get_iter(path)

        if index == -1:  # parent
            self.status = status
            iter = model.get_iter(path)
        else:
            iter = model.iter_nth_child(iter, index)

        model[iter][2] = int(status)

        # When the operation succeeds, ensure that the progressbars go
        # to 100 %, which is necessary when the operation doesnt
        # do any progress updated (which would be unfortunate!)
        if status == FileStatus.SUCCESS:
            model[iter][4] = 100.0
            model[iter][5] = "100.0 %"
        elif status == FileStatus.FAILURE:
            model[iter][6] = "red"
            model[iter][7] = GLib.markup_escape_text(message)
        elif status == FileStatus.SKIPPED:
            model[iter][6] = "grey"
            model[iter][7] = GLib.markup_escape_text(message)

        return GLib.SOURCE_REMOVE

    def update_status(
        self, index: int, status: FileStatus, message: Optional[str] = None
    ):
        """
        When an operation has finished, update the status of the corresponding
        entry in the treemodel.
        An index of -1 refers to the parent entry, 0 or higher refers to a child.
        """
        GLib.idle_add(self._update_status_worker_cb, index, status, message)

    def update_progressbar(self, index: int, value: float):
        """
        This method will update the progressbar of the current operation,
        defined by index, as well as the global one.
        value must be between 0 and 100.

        Try not to use this function too often, as it may slow the GUI
        down considerably. I recommend to use it only when value is a whole number
        """
        GLib.idle_add(self._update_progressbar_worker_cb, index, value)


class RegularFile(File):
    def __init__(
        self,
        filename: str,
        relative_filename: PurePath,
        created: int,
        status: FileStatus,
    ):

        super().__init__(
            filename,
            relative_filename,
            created,
            status,
        )


class WeightedRegularFile(RegularFile):
    def __init__(
        self,
        filename: str,
        relative_filename: PurePath,
        created: int,
        status: FileStatus,
        offset: float,  # fractional
        weight: float,  # fractional
    ):

        super().__init__(
            filename,
            relative_filename,
            created,
            status,
        )

        self._offset = offset
        self._weight = weight

    def update_progressbar(self, index: int, value: float):
        new_value = 100.0 * self._offset + value * self._weight
        GLib.idle_add(self._update_progressbar_worker_cb, index, new_value)


class Directory(File):
    def __init__(
        self,
        filename: str,
        relative_filename: PurePath,
        created: int,
        status: FileStatus,
        included_patterns: List[str],
        excluded_patterns: List[str],
    ):

        super().__init__(
            filename,
            relative_filename,
            created,
            status,
        )

        self._included_patterns = included_patterns
        self._excluded_patterns = excluded_patterns

        self._filelist: List[Tuple[str, int]] = []
        self._filelist_timestamp: int = 0
        self._total_size: int = 0

    @property
    def included_patterns(self):
        return self._included_patterns

    @property
    def excluded_patterns(self):
        return self._excluded_patterns

    def _get_filelist(self, _dir: Path) -> List[Tuple[str, int]]:
        rv: List[Tuple[str, int]] = []
        for entry in _dir.iterdir():
            if not match_path(
                entry,
                included_patterns=self._included_patterns,
                excluded_patterns=self._excluded_patterns,
                case_sensitive=False,
            ):
                continue
            if entry.is_file() and not entry.is_symlink():
                size = entry.stat().st_size
                self._total_size += size
                rv.append(
                    (
                        str(entry),
                        size,
                    )
                )
            elif entry.is_dir() and not entry.is_symlink():
                rv.extend(self._get_filelist(entry))
        return rv

    def _refresh_filelist(self):
        self._total_size = 0
        self._filelist = self._get_filelist(Path(self.filename))
        self._filelist_timestamp = time()

    @property
    def total_size(self):
        if self._filelist_timestamp < self._saved:
            self._refresh_filelist()
        return self._total_size

    def __iter__(self):
        if self._filelist_timestamp < self._saved:
            self._refresh_filelist()
        yield from self._filelist

    def __len__(self):
        return len(self._filelist)


class URL(File):
    def __init__(
        self,
        filename: str,
        relative_filename: PurePath,
        created: int,
        status: FileStatus,
    ):

        super().__init__(
            filename,
            relative_filename,
            created,
            status,
        )


class S3Object(File):
    def __init__(
        self,
        filename: str,
        relative_filename: PurePath,
        created: int,
        status: FileStatus,
        bucket_name: str,
        etag: str,
        size: int,
        region_name: str = "",
    ):

        super().__init__(
            filename,
            relative_filename,
            created,
            status,
        )
        self._bucket_name = bucket_name
        self._etag = etag
        self._size = size
        self._key = str(PurePosixPath(*self._relative_filename.parts))
        self._region_name = region_name

    @property
    def bucket_name(self):
        return self._bucket_name

    @property
    def etag(self):
        return self._etag

    @property
    def key(self):
        return self._key

    @property
    def size(self):
        return self._size

    @property
    def region_name(self):
        return self._region_name



from __future__ import annotations

import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, GLib
from ..utils import match_path
from watchdog.observers import Observer
from watchdog.events import PatternMatchingEventHandler

from ..engine import Engine, EngineThread
from ..utils import (
    LongTaskWindow,
    get_file_creation_timestamp,
    _get_common_patterns,
)
from ..file import FileStatus, RegularFile
from ..utils.decorators import (
    exported_filetype,
    with_advanced_settings,
    with_pango_docs,
)
from .file_watchdog_engine_advanced_settings import (
    FileWatchdogEngineAdvancedSettings,
)

from typing import List
from pathlib import Path, PurePath
import logging
import os

logger = logging.getLogger(__name__)

ERROR_MSG = "Ensure that the selected directory is readable and that any provided patterns do not conflict"


@with_pango_docs(filename="file_watchdog_engine.pango")
@with_advanced_settings(
    engine_advanced_settings=FileWatchdogEngineAdvancedSettings
)
@exported_filetype(filetype=RegularFile)
class FileWatchdogEngine(Engine):

    NAME = "Files Monitor"

    def __init__(self, appwindow):
        super().__init__(appwindow, FileWatchdogEngineThread, ERROR_MSG)

        label = Gtk.Label(
            label="Monitored Directory",
            halign=Gtk.Align.START,
            valign=Gtk.Align.CENTER,
            hexpand=False,
            vexpand=False,
        )
        self.attach(label, 0, 0, 1, 1)

        self._directory_chooser_button = self.register_widget(
            Gtk.FileChooserButton(
                title="Select a directory for monitoring",
                action=Gtk.FileChooserAction.SELECT_FOLDER,
                create_folders=True,
                halign=Gtk.Align.FILL,
                valign=Gtk.Align.FILL,
                hexpand=True,
                vexpand=False,
            ),
            "monitored_directory",
        )
        self.attach(self._directory_chooser_button, 1, 0, 1, 1)
        self._directory_chooser_button.connect(
            "selection-changed", self._directory_chooser_button_cb
        )

        self._monitor: Observer = None

    def _directory_chooser_button_cb(self, button):
        if (
            self.params.monitored_directory is None
            or Path(self.params.monitored_directory).is_dir() is False
        ):
            self._valid = False
        else:
            try:
                os.listdir(self.params.monitored_directory)
                self._valid = True
            except Exception:
                self._valid = False

        self.notify("valid")


# add our ExitableThread methods to Watchdog's Observer class to ensure it can be used as an EngineThread
class FileWatchdogEngineThread(Observer):
    def __init__(self, engine: FileWatchdogEngine, task_window: LongTaskWindow):
        super().__init__()
        self._should_exit: bool = False
        self._engine: Engine = engine
        self.params = engine.params
        self._task_window = task_window
        app = engine.appwindow.props.application
        self._included_patterns = app.get_allowed_file_patterns(
            self.params.allowed_patterns
        )
        self._excluded_patterns = app.get_ignored_file_patterns(
            self.params.ignore_patterns
        )

        self.schedule(
            event_handler=EventHandler(engine),
            path=engine.params.monitored_directory,
            recursive=engine.params.monitor_recursively,
        )

    @property
    def should_exit(self):
        return self._should_exit

    @should_exit.setter
    def should_exit(self, value: bool):
        self._should_exit = value
        if self._should_exit:
            self.stop()

    def _search_for_existing_files(self, directory: Path) -> List[RegularFile]:
        rv: List[RegularFile] = list()
        for child in directory.iterdir():
            if (
                child.is_file()
                and not child.is_symlink()
                and match_path(
                    child,
                    included_patterns=self._included_patterns,
                    excluded_patterns=self._excluded_patterns,
                    case_sensitive=False,
                )
            ):

                file_path = directory.joinpath(child)
                relative_file_path = file_path.relative_to(
                    self.params.monitored_directory
                )
                _file = RegularFile(
                    str(file_path),
                    relative_file_path,
                    get_file_creation_timestamp(file_path),
                    FileStatus.SAVED,
                )
                rv.append(_file)
            elif (
                self.params.monitor_recursively
                and child.is_dir()
                and not child.is_symlink()
            ):
                rv.extend(
                    self._search_for_existing_files(directory.joinpath(child))
                )
        return rv

    def run(self):
        # confirm patterns are valid
        if bool(
            common_patterns := _get_common_patterns(
                self._included_patterns, self._excluded_patterns, False
            )
        ):
            self._engine.cleanup()
            GLib.idle_add(
                self._engine.abort,
                self._task_window,
                f"Common patterns {common_patterns} detected!",
                priority=GLib.PRIORITY_HIGH,
            )
            return

        # check for existing files if necessary
        if self.params.process_existing_files:
            GLib.idle_add(
                self._task_window.set_text,
                "<b>Processing existing files...</b>",
            )
            try:
                existing_files = self._search_for_existing_files(
                    Path(self.params.monitored_directory)
                )
                GLib.idle_add(
                    self._engine._appwindow._queue_manager.add,
                    existing_files,
                    priority=GLib.PRIORITY_DEFAULT_IDLE,
                )
            except Exception as e:
                self._engine.cleanup()
                GLib.idle_add(
                    self._engine.abort,
                    self._task_window,
                    e,
                    priority=GLib.PRIORITY_HIGH,
                )
                return

        # if we get here, things should be working.
        # close task_window
        GLib.idle_add(
            self._engine.kill_task_window,
            self._task_window,
            priority=GLib.PRIORITY_HIGH,
        )

        # kick off watchdog's run method
        super().run()

    def on_thread_stop(self):
        super().on_thread_stop()
        self._engine.cleanup()


EngineThread.register(FileWatchdogEngineThread)


class EventHandler(PatternMatchingEventHandler):
    def __init__(self, engine: FileWatchdogEngine):
        self._engine = engine
        self.params = engine.params
        app = engine.appwindow.props.application
        included_patterns = app.get_allowed_file_patterns(
            self.params.allowed_patterns
        )
        ignore_patterns = app.get_ignored_file_patterns(
            self.params.ignore_patterns
        )
        super().__init__(
            patterns=included_patterns,
            ignore_patterns=ignore_patterns,
            ignore_directories=True,
        )

    def on_created(self, event):
        file_path = event.src_path
        logger.info(f"Monitor found {file_path} for event type CREATED")
        relative_file_path = PurePath(
            os.path.relpath(file_path, self.params.monitored_directory)
        )
        if (
            self._engine.props.running
            and self._engine._appwindow._queue_manager.props.running
        ):
            creation_timestamp = get_file_creation_timestamp(file_path)
            if creation_timestamp:
                _file = RegularFile(
                    file_path,
                    relative_file_path,
                    creation_timestamp,
                    FileStatus.CREATED,
                )
            else:
                logger.info(f"File Not found, {file_path} has been skipped")
                return None
            GLib.idle_add(
                self._engine._appwindow._queue_manager.add,
                _file,
                priority=GLib.PRIORITY_HIGH,
            )

    def on_modified(self, event):
        file_path = event.src_path
        logger.info(f"Monitor found {file_path} for event type MODIFIED")
        if (
            self._engine.props.running
            and self._engine._appwindow._queue_manager.props.running
        ):
            GLib.idle_add(
                self._engine._appwindow._queue_manager.saved,
                file_path,
                priority=GLib.PRIORITY_HIGH,
            )



from __future__ import annotations

import gi

gi.require_version("Gtk", "3.0")
from gi.repository import GLib, Gio, Gtk, GdkPixbuf
import yaml
from munch import Munch

import importlib.resources
import platform
import webbrowser
import logging
from typing import Dict, Type, Union, List, Optional
import importlib.metadata
from pathlib import Path

from .version import __version__
from .utils import (
    add_action_entries,
    PREFERENCES_CONFIG_FILE,
    MONITOR_YAML_VERSION,
)
from .preferences import (
    Preferences,
    AllowedFilePatternsPreference,
    IgnoredFilePatternsPreference,
)
from .preferenceswindow import PreferencesWindow
from .file import RegularFile, File
from .utils.helpwindow import HelpWindow
from .utils.googleanalytics import GoogleAnalyticsContext
from .applicationwindow import ApplicationWindow
from .engine import Engine
from .engine_advanced_settings import EngineAdvancedSettings
from .operation import Operation
from .queue_manager import QueueManager


logger = logging.getLogger(__name__)


class Application(Gtk.Application):
    def __init__(self, *args, **kwargs):
        super().__init__(
            *args,
            application_id="uk.ac.rfi.ai.file-monitor",
            flags=Gio.ApplicationFlags.FLAGS_NONE,
            **kwargs,
        )
        GLib.set_application_name("RFI File Monitor")

    @property
    def known_operations(self):
        return self._known_operations

    @property
    def known_engines(self):
        return self._known_engines

    @property
    def help_window(self):
        return self._help_window

    @property
    def engines_advanced_settings_map(self):
        return self._engines_advanced_settings_map

    @property
    def engines_exported_filetype_map(self):
        return self._engines_exported_filetype_map

    @property
    def filetypes_supported_operations_map(
        self,
    ) -> Dict[Type[File], List[Type[Operation]]]:
        return self._filetypes_supported_operations_map

    @property
    def pango_docs_map(self):
        return self._pango_docs_map

    @property
    def google_analytics_context(self):
        return self._google_analytics_context

    def do_shutdown(self):
        Gtk.Application.do_shutdown(self)

        self._google_analytics_context.consumer_thread.should_exit = True

    def do_startup(self):
        Gtk.Application.do_startup(self)

        # this may need to be checked on other platforms as well
        if platform.system() == "Darwin":
            appmenus_str = importlib.resources.read_text(
                "rfi_file_monitor.data", "menus-appmenu.ui"
            )
            builder = Gtk.Builder.new_from_string(appmenus_str, -1)
            self.set_app_menu(builder.get_object("app-menu"))

        commonmenus_str = importlib.resources.read_text(
            "rfi_file_monitor.data", "menus-common.ui"
        )
        builder = Gtk.Builder.new_from_string(commonmenus_str, -1)
        self.set_menubar(builder.get_object("menubar"))

        # get file filter menu
        popover_filter_menu_str = importlib.resources.read_text(
            "rfi_file_monitor.data", "filter-popover.ui"
        )
        builder = Gtk.Builder.new_from_string(popover_filter_menu_str, -1)
        self.filter_popover_menu = builder.get_object("filter-popover-menu")

        action_entries = (
            ("about", self.on_about),
            ("quit", self.on_quit),
            ("open", self.on_open),
            ("new", lambda *_: self.do_activate()),
            ("help-url", self.on_help_url, "s"),
            ("preferences", self.on_preferences),
        )

        # This doesn't work, which is kind of uncool
        # self.add_action_entries(action_entries)
        for action_entry in action_entries:
            add_action_entries(self, *action_entry)

        # add accelerators
        accelerators = (
            ("app.quit", ("<Primary>Q",)),
            ("app.new", ("<Primary>N",)),
            ("app.open", ("<Primary>O",)),
            ("win.save", ("<Primary>S",)),
            ("win.save-as", ("<Primary><Shift>S",)),
            ("win.close", ("<Primary>W",)),
        )

        for accel in accelerators:
            self.set_accels_for_action(accel[0], accel[1])

        self._engines_advanced_settings_map: Dict[
            Type[Engine], Type[EngineAdvancedSettings]
        ] = dict()

        self._engines_exported_filetype_map: Dict[
            Type[Engine], Type[File]
        ] = dict()

        self._filetypes_supported_operations_map: Dict[
            Type[File], List[Type[Operation]]
        ] = dict()

        self._pango_docs_map: Dict[
            Type[Union[Engine, QueueManager, Operation]], str
        ] = dict()

        # add queue manager docs manually
        try:
            contents = (
                Path(__file__)
                .parent.joinpath("docs", "queue_manager.pango")
                .read_text()
            )
        except Exception:
            logger.exception(
                f"with_pango_docs: could not open queue_manager.pango for reading"
            )
        else:
            self._pango_docs_map[QueueManager] = contents

        # get info from entry points
        self._known_operations = {
            e.name: e.load()
            for e in importlib.metadata.entry_points()[
                "rfi_file_monitor.operations"
            ]
        }

        self._update_supported_filetypes()

        self._known_engines = {
            e.name: e.load()
            for e in importlib.metadata.entry_points()[
                "rfi_file_monitor.engines"
            ]
        }

        # add our help window, which will be shared by all appwindows
        self._help_window = HelpWindow(self._pango_docs_map)

        # populate dict with preferences found in entry points
        self._prefs = Preferences(Munch(), Munch(), Munch())
        if "rfi_file_monitor.preferences" in importlib.metadata.entry_points():
            for e in importlib.metadata.entry_points()[
                "rfi_file_monitor.preferences"
            ]:
                _pref = e.load()
                self._prefs.settings[_pref] = _pref.default

        for _op in self._known_operations.values():
            self._prefs.operations[_op] = not bool(getattr(_op, "DEBUG", False))

        for _engine in self._known_engines.values():
            self._prefs.engines[_engine] = not bool(
                getattr(_engine, "DEBUG", False)
            )

        # now, open preferences file and update the prefs dictionary
        try:
            with PREFERENCES_CONFIG_FILE.open("r") as f:
                stored_prefs = yaml.safe_load(f)
        except FileNotFoundError:
            pass
        else:
            if stored_prefs and isinstance(stored_prefs, dict):
                # to maintain compatibility with older versions, first look for settings dict
                if (
                    "settings" in stored_prefs
                    and stored_prefs["settings"] is not None
                    and isinstance(stored_prefs["settings"], dict)
                ):
                    for _key, _value in stored_prefs["settings"].items():
                        for _pref in self._prefs.settings:
                            if _pref.key == _key:
                                self._prefs.settings[_pref] = _value
                                break
                        else:
                            logger.warning(
                                f"Could not find a corresponding Preference class for key {_key} from preferences file"
                            )
                else:
                    for _key, _value in stored_prefs.items():
                        for _pref in self._prefs.settings:
                            if _pref.key == _key:
                                self._prefs.settings[_pref] = _value
                                break
                        else:
                            logger.warning(
                                f"Could not find a corresponding Preference class for key {_key} from preferences file"
                            )

                if (
                    "operations" in stored_prefs
                    and stored_prefs["operations"] is not None
                    and isinstance(stored_prefs["operations"], dict)
                ):
                    for _key, _value in stored_prefs["operations"].items():
                        for _pref in self._prefs.operations:
                            if _pref.NAME == _key:
                                self._prefs.operations[_pref] = _value
                                break
                        else:
                            logger.warning(
                                f"Could not find a corresponding Operation class for key {_key} from preferences file"
                            )

                if (
                    "engines" in stored_prefs
                    and stored_prefs["engines"] is not None
                    and isinstance(stored_prefs["engines"], dict)
                ):
                    for _key, _value in stored_prefs["engines"].items():
                        for _pref in self._prefs.engines:
                            if _pref.NAME == _key:
                                self._prefs.engines[_pref] = _value
                                break
                        else:
                            logger.warning(
                                f"Could not find a corresponding Engine class for key {_key} from preferences file"
                            )

        # acquire google analytics context
        self._google_analytics_context = GoogleAnalyticsContext(
            endpoint="https://www.google-analytics.com/collect",
            tracking_id="UA-184737687-1",
            application_name="RFI-File-Monitor",
            application_version=__version__,
            config_file=Path(
                GLib.get_user_config_dir(), "rfi-file-monitor", "ga.conf"
            ),
        )

        # send event to Google Analytics
        self._google_analytics_context.send_event(
            "LAUNCH",
            "Monitor-{}-Python-{}-{}".format(
                __version__, platform.python_version(), platform.platform()
            ),
            None,
        )

    def _update_supported_filetypes(self):
        # this will update filetypes_supported_operations_map
        # with operations that were not decorated.
        # We will assume they support RegularFile only
        decorated_operations = list()
        for operations in self._filetypes_supported_operations_map.values():
            decorated_operations.extend(operations)
        decorated_operations = set(decorated_operations)

        undecorated_operations = set(
            self._known_operations.values()
        ).difference(decorated_operations)

        if RegularFile in self._filetypes_supported_operations_map:
            self._filetypes_supported_operations_map[RegularFile].extend(
                undecorated_operations
            )
        else:
            self._filetypes_supported_operations_map[RegularFile] = list(
                undecorated_operations
            )

        for operations in self._filetypes_supported_operations_map.values():
            operations.sort(key=lambda operation: operation.NAME)

    def get_preferences(self) -> Preferences:
        return self._prefs

    @classmethod
    def _split_patterns(cls, patterns: Optional[str]) -> List[str]:
        if patterns and patterns.split():
            return list(map(lambda x: x.strip(), patterns.split(",")))
        return []

    def get_allowed_file_patterns(
        self, extra_patterns: Optional[str] = None
    ) -> List[str]:
        global_allowed_patterns: str = self._prefs.settings[
            AllowedFilePatternsPreference
        ]
        patterns = self._split_patterns(global_allowed_patterns)
        patterns.extend(self._split_patterns(extra_patterns))

        # if no patterns were found, match everything!
        if not patterns:
            return ["*"]
        return patterns

    def get_ignored_file_patterns(
        self, extra_patterns: Optional[str] = None
    ) -> List[str]:
        global_ignored_patterns: str = self._prefs.settings[
            IgnoredFilePatternsPreference
        ]
        patterns = self._split_patterns(global_ignored_patterns)
        patterns.extend(self._split_patterns(extra_patterns))
        return patterns

    def on_open(self, action, param):
        # fire up file chooser dialog
        active_window = self.get_active_window()
        dialog = Gtk.FileChooserNative(
            modal=True,
            title="Open monitor configuration YAML file",
            transient_for=active_window,
            action=Gtk.FileChooserAction.OPEN,
        )
        filter = Gtk.FileFilter()
        filter.add_pattern("*.yml")
        filter.add_pattern("*.yaml")
        filter.set_name("YAML file")
        dialog.add_filter(filter)

        if dialog.run() == Gtk.ResponseType.ACCEPT:
            yaml_file = dialog.get_filename()
            dialog.destroy()
            try:
                with open(yaml_file, "r") as f:
                    yaml_dict = yaml.safe_load(f)
                if (
                    "version" not in yaml_dict
                    or yaml_dict["version"] != MONITOR_YAML_VERSION
                ):
                    raise Exception(
                        f"The YAML file {yaml_file} is not compatible with this version of the RFI-File-Monitor"
                    )
                if (
                    "active_engine" not in yaml_dict
                    or "queue_manager" not in yaml_dict
                    or "operations" not in yaml_dict
                    or "engines" not in yaml_dict
                ):
                    raise Exception(
                        f"The YAML file {yaml_file} is incomplete and cannot be loaded"
                    )
            except Exception as e:
                dialog = Gtk.MessageDialog(
                    transient_for=active_window,
                    modal=True,
                    destroy_with_parent=True,
                    message_type=Gtk.MessageType.ERROR,
                    buttons=Gtk.ButtonsType.CLOSE,
                    text=f"Could not load {yaml_file}",
                    secondary_text=str(e),
                )
                dialog.run()
                dialog.destroy()
            else:
                window = ApplicationWindow(
                    application=self,
                    type=Gtk.WindowType.TOPLEVEL,
                    force_all=True,
                )
                window.show_all()
                window.load_from_yaml_dict(yaml_dict)
        else:
            dialog.destroy()

    def do_activate(self):
        window = ApplicationWindow(
            application=self,
            title="RFI-File-Monitor",
            type=Gtk.WindowType.TOPLEVEL,
        )
        window.show_all()

    def on_about(self, action, param):

        with importlib.resources.path(
            "rfi_file_monitor.data", "RFI-logo-transparent.png"
        ) as f:
            logo = GdkPixbuf.Pixbuf.new_from_file_at_scale(
                str(f), 300, -1, True
            )

        about_dialog = Gtk.AboutDialog(
            transient_for=self.get_active_window(),
            modal=True,
            authors=["Tom Schoonjans"],
            logo=logo,
            version=__version__,
        )
        about_dialog.present()

    def on_quit(self, action, param):
        windows = filter(
            lambda window: isinstance(window, ApplicationWindow),
            self.get_windows(),
        )

        for index, window in enumerate(windows):
            if window.active_engine.props.running:
                window.close()
            else:
                self.remove_window(window)

    def on_help_url(self, action, param):
        webbrowser.open_new_tab(param.get_string())

    def on_preferences(self, action, param):
        window = PreferencesWindow(
            self._prefs,
            modal=False,
            transient_for=self.get_active_window(),
            window_position=Gtk.WindowPosition.CENTER_ON_PARENT,
            type=Gtk.WindowType.TOPLEVEL,
            destroy_with_parent=True,
            border_width=5,
        )
        window.present()
